import numpy as np
from desc.backend import jnp, put
from scipy.constants import mu_0

"""These functions perform the core calculations of physical quantities.
They are used as methods of the Configuration class, and also used to compute
quantities in the objective functions.
All of the functions in this file have the same call signature:

Parameters
----------
Psi : float
    total toroidal flux (in Webers) within the last closed flux surface
R_lmn : ndarray
    spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
Z_lmn : ndarray
    spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
L_lmn : ndarray
    spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
p_l : ndarray
    spectral coefficients of p(rho) -- pressure profile
i_l : ndarray
    spectral coefficients of iota(rho) -- rotational transform profile
R_transform : Transform
    transforms R_lmn coefficients to real space
Z_transform : Transform
    transforms Z_lmn coefficients to real space
L_transform : Transform
    transforms L_lmn coefficients to real space
p_transform : Transform
    transforms p_l coefficients to real space
i_transform : Transform
    transforms i_l coefficients to real space
zeta_ratio : float
    scale factor for zeta derivatives. Setting to zero effectively solves
    for individual tokamak solutions at each toroidal plane,
    setting to 1 solves for a stellarator. (Default value = 1.0)

"""


def dot(a, b, axis):
    """Batched vector dot product.

    Parameters
    ----------
    a : array-like
        first array of vectors
    b : array-like
        second array of vectors
    axis : int
        axis along which vectors are stored

    Returns
    -------
    y : array-like
        y = sum(a*b, axis=axis)

    """
    return jnp.sum(a * b, axis=axis, keepdims=False)


def cross(a, b, axis):
    """Batched vector cross product.

    Parameters
    ----------
    a : array-like
        first array of vectors
    b : array-like
        second array of vectors
    axis : int
        axis along which vectors are stored

    Returns
    -------
    y : array-like
        y = a x b

    """
    return jnp.cross(a, b, axis=axis)


def compute_profiles(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute magnetic flux, pressure, and rotational transform profiles.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    profiles : dict
        dictionary of ndarray, shape(num_nodes,) of profiles.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    profiles = {}

    # toroidal flux
    rho = p_transform.grid.nodes[:, 0]
    profiles["psi"] = Psi * rho ** 2
    profiles["psi_r"] = 2 * Psi * rho
    profiles["psi_rr"] = 2 * Psi * np.ones_like(rho)

    # pressure
    profiles["p"] = p_transform.transform(p_l, 0)
    profiles["p_r"] = p_transform.transform(p_l, 1)

    # rotational transform
    profiles["iota"] = i_transform.transform(i_l, 0)
    profiles["iota_r"] = i_transform.transform(i_l, 1)

    return profiles


def compute_toroidal_coords(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Transform toroidal coordinates to real space.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    coordinates : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    toroidal_coords = {}
    toroidal_coords["R"] = R_transform.transform(R_lmn)
    toroidal_coords["Z"] = Z_transform.transform(Z_lmn)
    toroidal_coords["lambda"] = L_transform.transform(L_lmn)
    toroidal_coords["0"] = jnp.zeros_like(toroidal_coords["R"])

    return toroidal_coords


def compute_cartesian_coords(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute cartesian coordinates from toroidal coordinates.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    cartesian_coords : dict
        dictionary of ndarray, shape(num_nodes,) of cartesian coordinates
        evaluated at grid nodes.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    toroidal_coords = compute_toroidal_coords(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    cartesian_coords = {}
    phi = R_transform.grid.nodes[:, 2]
    cartesian_coords["X"] = toroidal_coords["R"] * np.cos(phi)
    cartesian_coords["Y"] = toroidal_coords["R"] * np.sin(phi)
    cartesian_coords["Z"] = toroidal_coords["Z"]

    return cartesian_coords, toroidal_coords


def compute_covariant_basis(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute covariant basis vectors.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    cov_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of covariant basis vectors.
        Keys are of the form 'e_x_y', meaning the covariant basis vector in
        the x direction, differentiated wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    toroidal_coords = compute_toroidal_coords(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    # toroidal coordinate 1st derivatives
    toroidal_coords["R_r"] = R_transform.transform(R_lmn, 1, 0, 0)
    toroidal_coords["Z_r"] = Z_transform.transform(Z_lmn, 1, 0, 0)
    toroidal_coords["R_t"] = R_transform.transform(R_lmn, 0, 1, 0)
    toroidal_coords["Z_t"] = Z_transform.transform(Z_lmn, 0, 1, 0)
    toroidal_coords["R_z"] = R_transform.transform(R_lmn, 0, 0, 1) * zeta_ratio
    toroidal_coords["Z_z"] = Z_transform.transform(Z_lmn, 0, 0, 1) * zeta_ratio

    cov_basis = {}
    cov_basis["e_rho"] = jnp.array(
        [toroidal_coords["R_r"], toroidal_coords["0"], toroidal_coords["Z_r"]]
    )
    cov_basis["e_theta"] = jnp.array(
        [toroidal_coords["R_t"], toroidal_coords["0"], toroidal_coords["Z_t"]]
    )
    cov_basis["e_zeta"] = jnp.array(
        [toroidal_coords["R_z"], toroidal_coords["R"], toroidal_coords["Z_z"]]
    )

    return cov_basis, toroidal_coords


def compute_jacobian(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute coordinate system jacobian.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    jacobian : dict
        dictionary of ndarray, shape(num_nodes,), of coordinate system jacobian.
        Keys are of the form 'g_x' meaning the x derivative of the coordinate
        system jacobian g.
    cov_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of covariant basis vectors.
        Keys are of the form 'e_x_y', meaning the covariant basis vector in
        the x direction, differentiated wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    cov_basis, toroidal_coords = compute_covariant_basis(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    jacobian = {}
    jacobian["g"] = dot(
        cov_basis["e_rho"], cross(cov_basis["e_theta"], cov_basis["e_zeta"], 0), 0
    )

    return jacobian, cov_basis, toroidal_coords


def compute_contravariant_basis(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute contravariant basis vectors.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    con_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of contravariant basis vectors.
        Keys are of the form 'e^x_y', meaning the contravariant basis vector
        in the x direction, differentiated wrt to y.
    jacobian : dict
        dictionary of ndarray, shape(num_nodes,), of coordinate system jacobian.
        Keys are of the form 'g_x' meaning the x derivative of the coordinate
        system jacobian g.
    cov_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of covariant basis vectors.
        Keys are of the form 'e_x_y', meaning the covariant basis vector in
        the x direction, differentiated wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    jacobian, cov_basis, toroidal_coords = compute_jacobian(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    con_basis = {}
    con_basis["e^rho"] = (
        cross(cov_basis["e_theta"], cov_basis["e_zeta"], 0) / jacobian["g"]
    )
    con_basis["e^theta"] = (
        cross(cov_basis["e_zeta"], cov_basis["e_rho"], 0) / jacobian["g"]
    )
    con_basis["e^zeta"] = jnp.array(
        [toroidal_coords["0"], 1 / toroidal_coords["R"], toroidal_coords["0"]]
    )

    return con_basis, jacobian, cov_basis, toroidal_coords


def compute_magnetic_field(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute magnetic field components.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    magnetic_field: dict
        dictionary of ndarray, shape(num_nodes,) of magnetic field components.
        Keys are of the form 'B_x_y' or 'B^x_y', meaning the covariant (B_x)
        or contravariant (B^x) component of the magnetic field, with the
        derivative wrt to y.
    jacobian : dict
        dictionary of ndarray, shape(num_nodes,), of coordinate system jacobian.
        Keys are of the form 'g_x' meaning the x derivative of the coordinate
        system jacobian g.
    cov_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of covariant basis vectors.
        Keys are of the form 'e_x_y', meaning the covariant basis vector in
        the x direction, differentiated wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.
    profiles : dict
        dictionary of ndarray, shape(num_nodes,) of profiles.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    profiles = compute_profiles(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )
    jacobian, cov_basis, toroidal_coords = compute_jacobian(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    # lambda derivatives
    toroidal_coords["lambda_t"] = L_transform.transform(L_lmn, 0, 1, 0)
    toroidal_coords["lambda_z"] = L_transform.transform(L_lmn, 0, 0, 1) * zeta_ratio

    magnetic_field = {}
    magnetic_field["B0"] = profiles["psi_r"] / (2 * jnp.pi * jacobian["g"])

    # contravariant components
    magnetic_field["B^rho"] = jnp.zeros_like(magnetic_field["B0"])
    magnetic_field["B^theta"] = magnetic_field["B0"] * (
        profiles["iota"] - toroidal_coords["lambda_z"]
    )
    magnetic_field["B^zeta"] = magnetic_field["B0"] * (1 + toroidal_coords["lambda_t"])
    magnetic_field["B"] = (
        magnetic_field["B^theta"] * cov_basis["e_theta"]
        + magnetic_field["B^zeta"] * cov_basis["e_zeta"]
    )

    # covariant components
    magnetic_field["B_rho"] = dot(magnetic_field["B"], cov_basis["e_rho"], 0)
    magnetic_field["B_theta"] = dot(magnetic_field["B"], cov_basis["e_theta"], 0)
    magnetic_field["B_zeta"] = dot(magnetic_field["B"], cov_basis["e_zeta"], 0)

    return magnetic_field, jacobian, cov_basis, toroidal_coords, profiles


def compute_magnetic_field_axis(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute magnetic field components; can handle nodes at the magnetic axis.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    magnetic_field: dict
        dictionary of ndarray, shape(num_nodes,) of magnetic field components.
        Keys are of the form 'B_x_y' or 'B^x_y', meaning the covariant (B_x)
        or contravariant (B^x) component of the magnetic field, with the
        derivative wrt to y.
    jacobian : dict
        dictionary of ndarray, shape(num_nodes,), of coordinate system jacobian.
        Keys are of the form 'g_x' meaning the x derivative of the coordinate
        system jacobian g.
    cov_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of covariant basis vectors.
        Keys are of the form 'e_x_y', meaning the covariant basis vector in
        the x direction, differentiated wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.
    profiles : dict
        dictionary of ndarray, shape(num_nodes,) of profiles.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    profiles = compute_profiles(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )
    jacobian, cov_basis, toroidal_coords = compute_jacobian(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    axis = i_transform.grid.axis

    # lambda derivatives
    toroidal_coords["lambda_t"] = L_transform.transform(L_lmn, 0, 1, 0)
    toroidal_coords["lambda_z"] = L_transform.transform(L_lmn, 0, 0, 1) * zeta_ratio

    # toroidal coordinate 2nd derivatives
    toroidal_coords["R_rr"] = R_transform.transform(R_lmn, 2, 0, 0)
    toroidal_coords["Z_rr"] = Z_transform.transform(Z_lmn, 2, 0, 0)
    toroidal_coords["R_rt"] = R_transform.transform(R_lmn, 1, 1, 0)
    toroidal_coords["Z_rt"] = Z_transform.transform(Z_lmn, 1, 1, 0)
    toroidal_coords["R_rz"] = R_transform.transform(R_lmn, 1, 0, 1)
    toroidal_coords["Z_rz"] = Z_transform.transform(Z_lmn, 1, 0, 1)

    # covariant basis derivatives
    cov_basis["e_rho_r"] = jnp.array(
        [toroidal_coords["R_rr"], toroidal_coords["0"], toroidal_coords["Z_rr"]]
    )
    cov_basis["e_theta_r"] = jnp.array(
        [toroidal_coords["R_rt"], toroidal_coords["0"], toroidal_coords["Z_rt"]]
    )
    cov_basis["e_zeta_r"] = jnp.array(
        [toroidal_coords["R_rz"], toroidal_coords["R_r"], toroidal_coords["Z_rz"]]
    )

    # jacobian derivatives
    jacobian["g_r"] = (
        dot(
            cov_basis["e_rho_r"], cross(cov_basis["e_theta"], cov_basis["e_zeta"], 0), 0
        )
        + dot(
            cov_basis["e_rho"], cross(cov_basis["e_theta_r"], cov_basis["e_zeta"], 0), 0
        )
        + dot(
            cov_basis["e_rho"], cross(cov_basis["e_theta"], cov_basis["e_zeta_r"], 0), 0
        )
    )

    magnetic_field = {}
    magnetic_field["B0"] = profiles["psi_r"] / (2 * jnp.pi * jacobian["g"])
    magnetic_field["B0"] = put(
        magnetic_field["B0"],
        axis,
        profiles["psi_rr"][axis] / (2 * jnp.pi * jacobian["g_r"][axis]),
    )

    # contravariant components
    magnetic_field["B^rho"] = jnp.zeros_like(magnetic_field["B0"])
    magnetic_field["B^theta"] = magnetic_field["B0"] * (
        profiles["iota"] - toroidal_coords["lambda_z"]
    )
    magnetic_field["B^zeta"] = magnetic_field["B0"] * (1 + toroidal_coords["lambda_t"])
    magnetic_field["B"] = (
        magnetic_field["B^theta"] * cov_basis["e_theta"]
        + magnetic_field["B^zeta"] * cov_basis["e_zeta"]
    )

    # covariant components
    magnetic_field["B_rho"] = dot(magnetic_field["B"], cov_basis["e_rho"], 0)
    magnetic_field["B_theta"] = dot(magnetic_field["B"], cov_basis["e_theta"], 0)
    magnetic_field["B_zeta"] = dot(magnetic_field["B"], cov_basis["e_zeta"], 0)

    return magnetic_field, jacobian, cov_basis, toroidal_coords, profiles


def compute_magnetic_field_magnitude(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute magnetic field magnitude.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    magnetic_field: dict
        dictionary of ndarray, shape(num_nodes,) of magnetic field components.
        Keys are of the form 'B_x_y' or 'B^x_y', meaning the covariant (B_x)
        or contravariant (B^x) component of the magnetic field, with the
        derivative wrt to y.
    jacobian : dict
        dictionary of ndarray, shape(num_nodes,), of coordinate system jacobian.
        Keys are of the form 'g_x' meaning the x derivative of the coordinate
        system jacobian g.
    cov_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of covariant basis vectors.
        Keys are of the form 'e_x_y', meaning the covariant basis vector in
        the x direction, differentiated wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.
    profiles : dict
        dictionary of ndarray, shape(num_nodes,) of profiles.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    (
        magnetic_field,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    ) = compute_magnetic_field(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    magnetic_field["|B|"] = jnp.sqrt(
        magnetic_field["B^theta"] ** 2
        * dot(cov_basis["e_theta"], cov_basis["e_theta"], 0)
        + magnetic_field["B^zeta"] ** 2
        * dot(cov_basis["e_zeta"], cov_basis["e_zeta"], 0)
        + 2
        * magnetic_field["B^theta"]
        * magnetic_field["B^zeta"]
        * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
    )

    return magnetic_field, jacobian, cov_basis, toroidal_coords, profiles


def compute_magnetic_field_magnitude_axis(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute magnetic field magnitude; can handle nodes at the magnetic axis.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    magnetic_field: dict
        dictionary of ndarray, shape(num_nodes,) of magnetic field components.
        Keys are of the form 'B_x_y' or 'B^x_y', meaning the covariant (B_x)
        or contravariant (B^x) component of the magnetic field, with the
        derivative wrt to y.
    jacobian : dict
        dictionary of ndarray, shape(num_nodes,), of coordinate system jacobian.
        Keys are of the form 'g_x' meaning the x derivative of the coordinate
        system jacobian g.
    cov_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of covariant basis vectors.
        Keys are of the form 'e_x_y', meaning the covariant basis vector in
        the x direction, differentiated wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.
    profiles : dict
        dictionary of ndarray, shape(num_nodes,) of profiles.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    (
        magnetic_field,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    ) = compute_magnetic_field_axis(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    magnetic_field["|B|"] = jnp.sqrt(
        magnetic_field["B^theta"] ** 2
        * dot(cov_basis["e_theta"], cov_basis["e_theta"], 0)
        + magnetic_field["B^zeta"] ** 2
        * dot(cov_basis["e_zeta"], cov_basis["e_zeta"], 0)
        + 2
        * magnetic_field["B^theta"]
        * magnetic_field["B^zeta"]
        * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
    )

    return magnetic_field, jacobian, cov_basis, toroidal_coords, profiles


def compute_current_density(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute current density field components.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    current_density : dict
        dictionary of ndarray, shape(num_nodes,), of current density components.
        Keys are of the form 'J^x_y' meaning the contravariant (J^x)
        component of the current, with the derivative wrt to y.
    magnetic_field: dict
        dictionary of ndarray, shape(num_nodes,) of magnetic field components.
        Keys are of the form 'B_x_y' or 'B^x_y', meaning the covariant (B_x)
        or contravariant (B^x) component of the magnetic field, with the
        derivative wrt to y.
    jacobian : dict
        dictionary of ndarray, shape(num_nodes,), of coordinate system jacobian.
        Keys are of the form 'g_x' meaning the x derivative of the coordinate
        system jacobian g.
    cov_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of covariant basis vectors.
        Keys are of the form 'e_x_y', meaning the covariant basis vector in
        the x direction, differentiated wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.
    profiles : dict
        dictionary of ndarray, shape(num_nodes,) of profiles.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    (
        magnetic_field,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    ) = compute_magnetic_field(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    # toroidal coordinate 2nd derivatives
    toroidal_coords["R_rr"] = R_transform.transform(R_lmn, 2, 0, 0)
    toroidal_coords["Z_rr"] = Z_transform.transform(Z_lmn, 2, 0, 0)
    toroidal_coords["R_rt"] = R_transform.transform(R_lmn, 1, 1, 0)
    toroidal_coords["Z_rt"] = Z_transform.transform(Z_lmn, 1, 1, 0)
    toroidal_coords["R_rz"] = R_transform.transform(R_lmn, 1, 0, 1) * zeta_ratio
    toroidal_coords["Z_rz"] = Z_transform.transform(Z_lmn, 1, 0, 1) * zeta_ratio
    toroidal_coords["R_tt"] = R_transform.transform(R_lmn, 0, 2, 0)
    toroidal_coords["Z_tt"] = Z_transform.transform(Z_lmn, 0, 2, 0)
    toroidal_coords["R_tz"] = R_transform.transform(R_lmn, 0, 1, 1) * zeta_ratio
    toroidal_coords["Z_tz"] = Z_transform.transform(Z_lmn, 0, 1, 1) * zeta_ratio
    toroidal_coords["R_zz"] = R_transform.transform(R_lmn, 0, 0, 2) * zeta_ratio
    toroidal_coords["Z_zz"] = Z_transform.transform(Z_lmn, 0, 0, 2) * zeta_ratio

    # lambda derivatives
    toroidal_coords["lambda_rt"] = L_transform.transform(L_lmn, 1, 1, 0)
    toroidal_coords["lambda_rz"] = L_transform.transform(L_lmn, 1, 0, 1) * zeta_ratio
    toroidal_coords["lambda_tt"] = L_transform.transform(L_lmn, 0, 2, 0)
    toroidal_coords["lambda_tz"] = L_transform.transform(L_lmn, 0, 1, 1) * zeta_ratio
    toroidal_coords["lambda_zz"] = L_transform.transform(L_lmn, 0, 0, 2) * zeta_ratio

    # covariant basis derivatives
    cov_basis["e_rho_r"] = jnp.array(
        [toroidal_coords["R_rr"], toroidal_coords["0"], toroidal_coords["Z_rr"]]
    )
    cov_basis["e_rho_t"] = jnp.array(
        [toroidal_coords["R_rt"], toroidal_coords["0"], toroidal_coords["Z_rt"]]
    )
    cov_basis["e_rho_z"] = jnp.array(
        [toroidal_coords["R_rz"], toroidal_coords["0"], toroidal_coords["Z_rz"]]
    )

    cov_basis["e_theta_r"] = jnp.array(
        [toroidal_coords["R_rt"], toroidal_coords["0"], toroidal_coords["Z_rt"]]
    )
    cov_basis["e_theta_t"] = jnp.array(
        [toroidal_coords["R_tt"], toroidal_coords["0"], toroidal_coords["Z_tt"]]
    )
    cov_basis["e_theta_z"] = jnp.array(
        [toroidal_coords["R_tz"], toroidal_coords["0"], toroidal_coords["Z_tz"]]
    )

    cov_basis["e_zeta_r"] = jnp.array(
        [toroidal_coords["R_rz"], toroidal_coords["R_r"], toroidal_coords["Z_rz"]]
    )
    cov_basis["e_zeta_t"] = jnp.array(
        [toroidal_coords["R_tz"], toroidal_coords["R_t"], toroidal_coords["Z_tz"]]
    )
    cov_basis["e_zeta_z"] = jnp.array(
        [toroidal_coords["R_zz"], toroidal_coords["R_z"], toroidal_coords["Z_zz"]]
    )

    # jacobian derivatives
    jacobian["g_r"] = (
        dot(
            cov_basis["e_rho_r"], cross(cov_basis["e_theta"], cov_basis["e_zeta"], 0), 0
        )
        + dot(
            cov_basis["e_rho"], cross(cov_basis["e_theta_r"], cov_basis["e_zeta"], 0), 0
        )
        + dot(
            cov_basis["e_rho"], cross(cov_basis["e_theta"], cov_basis["e_zeta_r"], 0), 0
        )
    )
    jacobian["g_t"] = (
        dot(
            cov_basis["e_rho_t"], cross(cov_basis["e_theta"], cov_basis["e_zeta"], 0), 0
        )
        + dot(
            cov_basis["e_rho"], cross(cov_basis["e_theta_t"], cov_basis["e_zeta"], 0), 0
        )
        + dot(
            cov_basis["e_rho"], cross(cov_basis["e_theta"], cov_basis["e_zeta_t"], 0), 0
        )
    )
    jacobian["g_z"] = (
        dot(
            cov_basis["e_rho_z"], cross(cov_basis["e_theta"], cov_basis["e_zeta"], 0), 0
        )
        + dot(
            cov_basis["e_rho"], cross(cov_basis["e_theta_z"], cov_basis["e_zeta"], 0), 0
        )
        + dot(
            cov_basis["e_rho"], cross(cov_basis["e_theta"], cov_basis["e_zeta_z"], 0), 0
        )
    )

    # B contravariant component derivatives
    magnetic_field["B0_r"] = profiles["psi_rr"] / (2 * jnp.pi * jacobian["g"]) - (
        profiles["psi_r"] * jacobian["g_r"]
    ) / (2 * jnp.pi * jacobian["g"] ** 2)
    magnetic_field["B0_t"] = -(profiles["psi_r"] * jacobian["g_t"]) / (
        2 * jnp.pi * jacobian["g"] ** 2
    )
    magnetic_field["B0_z"] = -(profiles["psi_r"] * jacobian["g_z"]) / (
        2 * jnp.pi * jacobian["g"] ** 2
    )
    magnetic_field["B^theta_r"] = magnetic_field["B0_r"] * (
        profiles["iota"] - toroidal_coords["lambda_z"]
    ) + magnetic_field["B0"] * (profiles["iota_r"] - toroidal_coords["lambda_rz"])
    magnetic_field["B^theta_t"] = (
        magnetic_field["B0_t"] * (profiles["iota"] - toroidal_coords["lambda_z"])
        - magnetic_field["B0"] * toroidal_coords["lambda_tz"]
    )
    magnetic_field["B^theta_z"] = (
        magnetic_field["B0_z"] * (profiles["iota"] - toroidal_coords["lambda_z"])
        - magnetic_field["B0"] * toroidal_coords["lambda_zz"]
    )
    magnetic_field["B^zeta_r"] = (
        magnetic_field["B0_r"] * (1 + toroidal_coords["lambda_t"])
        + magnetic_field["B0"] * toroidal_coords["lambda_rt"]
    )
    magnetic_field["B^zeta_t"] = (
        magnetic_field["B0_t"] * (1 + toroidal_coords["lambda_t"])
        + magnetic_field["B0"] * toroidal_coords["lambda_tt"]
    )
    magnetic_field["B^zeta_z"] = (
        magnetic_field["B0_z"] * (1 + toroidal_coords["lambda_t"])
        + magnetic_field["B0"] * toroidal_coords["lambda_tz"]
    )

    # B vector partial derivatives
    magnetic_field["B_r"] = (
        magnetic_field["B^theta_r"] * cov_basis["e_theta"]
        + magnetic_field["B^theta"] * cov_basis["e_theta_r"]
        + magnetic_field["B^zeta_r"] * cov_basis["e_zeta"]
        + magnetic_field["B^zeta"] * cov_basis["e_zeta_r"]
    )
    magnetic_field["B_t"] = (
        magnetic_field["B^theta_t"] * cov_basis["e_theta"]
        + magnetic_field["B^theta"] * cov_basis["e_theta_t"]
        + magnetic_field["B^zeta_t"] * cov_basis["e_zeta"]
        + magnetic_field["B^zeta"] * cov_basis["e_zeta_t"]
    )
    magnetic_field["B_z"] = (
        magnetic_field["B^theta_z"] * cov_basis["e_theta"]
        + magnetic_field["B^theta"] * cov_basis["e_theta_z"]
        + magnetic_field["B^zeta_z"] * cov_basis["e_zeta"]
        + magnetic_field["B^zeta"] * cov_basis["e_zeta_z"]
    )

    # B covariant component derivatives
    magnetic_field["B_rho_t"] = dot(magnetic_field["B_t"], cov_basis["e_rho"], 0) + dot(
        magnetic_field["B"], cov_basis["e_rho_t"], 0
    )
    magnetic_field["B_rho_z"] = dot(magnetic_field["B_z"], cov_basis["e_rho"], 0) + dot(
        magnetic_field["B"], cov_basis["e_rho_z"], 0
    )
    magnetic_field["B_theta_r"] = dot(
        magnetic_field["B_r"], cov_basis["e_theta"], 0
    ) + dot(magnetic_field["B"], cov_basis["e_theta_r"], 0)
    magnetic_field["B_theta_z"] = dot(
        magnetic_field["B_z"], cov_basis["e_theta"], 0
    ) + dot(magnetic_field["B"], cov_basis["e_theta_z"], 0)
    magnetic_field["B_zeta_r"] = dot(
        magnetic_field["B_r"], cov_basis["e_zeta"], 0
    ) + dot(magnetic_field["B"], cov_basis["e_zeta_r"], 0)
    magnetic_field["B_zeta_t"] = dot(
        magnetic_field["B_t"], cov_basis["e_zeta"], 0
    ) + dot(magnetic_field["B"], cov_basis["e_zeta_t"], 0)

    current_density = {}

    # J contravariant components
    current_density["J^rho"] = (
        magnetic_field["B_zeta_t"] - magnetic_field["B_theta_z"]
    ) / (mu_0 * jacobian["g"])
    current_density["J^theta"] = (
        magnetic_field["B_rho_z"] - magnetic_field["B_zeta_r"]
    ) / (mu_0 * jacobian["g"])
    current_density["J^zeta"] = (
        magnetic_field["B_theta_r"] - magnetic_field["B_rho_t"]
    ) / (mu_0 * jacobian["g"])

    return (
        current_density,
        magnetic_field,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    )


def compute_magnetic_pressure_gradient(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute magnetic pressure gradient components and its magnitude.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    magnetic_pressure : dict
        dictionary of ndarray, shape(num_nodes,), of magnetic pressure gradient components.
        Keys are of the form 'grad(B)^x' meaning the contravariant (grad(B)^x)
        component of the magnetic pressure gradient.
    current_density : dict
        dictionary of ndarray, shape(num_nodes,), of current density components.
        Keys are of the form 'J^x_y' meaning the contravariant (J^x)
        component of the current, with the derivative wrt to y.
    magnetic_field: dict
        dictionary of ndarray, shape(num_nodes,) of magnetic field components.
        Keys are of the form 'B_x_y' or 'B^x_y', meaning the covariant (B_x)
        or contravariant (B^x) component of the magnetic field, with the
        derivative wrt to y.
    con_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of contravariant basis vectors.
        Keys are of the form 'e^x_y', meaning the contravariant basis vector
        in the x direction, differentiated wrt to y.
    jacobian : dict
        dictionary of ndarray, shape(num_nodes,), of coordinate system jacobian.
        Keys are of the form 'g_x' meaning the x derivative of the coordinate
        system jacobian g.
    cov_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of covariant basis vectors.
        Keys are of the form 'e_x_y', meaning the covariant basis vector in
        the x direction, differentiated wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.
    profiles : dict
        dictionary of ndarray, shape(num_nodes,) of profiles.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    (
        current_density,
        magnetic_field,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    ) = compute_current_density(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    # contravariant basis vectors
    con_basis = {}
    con_basis["e^rho"] = (
        cross(cov_basis["e_theta"], cov_basis["e_zeta"], 0) / jacobian["g"]
    )
    con_basis["e^theta"] = (
        cross(cov_basis["e_zeta"], cov_basis["e_rho"], 0) / jacobian["g"]
    )
    con_basis["e^zeta"] = jnp.array(
        [toroidal_coords["0"], 1 / toroidal_coords["R"], toroidal_coords["0"]]
    )

    magnetic_pressure = {}

    # B covariant component derivatives
    magnetic_field["B_theta_t"] = dot(
        magnetic_field["B_t"], cov_basis["e_theta"], 0
    ) + dot(magnetic_field["B"], cov_basis["e_theta_t"], 0)
    magnetic_field["B_zeta_z"] = dot(
        magnetic_field["B_z"], cov_basis["e_zeta"], 0
    ) + dot(magnetic_field["B"], cov_basis["e_zeta_z"], 0)

    # magnetic pressure gradient covariant components
    magnetic_pressure["grad(|B|^2)_rho"] = (
        magnetic_field["B^theta"] * magnetic_field["B_theta_r"]
        + magnetic_field["B_theta"] * magnetic_field["B^theta_r"]
        + magnetic_field["B^zeta"] * magnetic_field["B_zeta_r"]
        + magnetic_field["B_zeta"] * magnetic_field["B^zeta_r"]
    )
    magnetic_pressure["grad(|B|^2)_theta"] = (
        magnetic_field["B^theta"] * magnetic_field["B_theta_t"]
        + magnetic_field["B_theta"] * magnetic_field["B^theta_t"]
        + magnetic_field["B^zeta"] * magnetic_field["B_zeta_t"]
        + magnetic_field["B_zeta"] * magnetic_field["B^zeta_t"]
    )
    magnetic_pressure["grad(|B|^2)_zeta"] = (
        magnetic_field["B^theta"] * magnetic_field["B_theta_z"]
        + magnetic_field["B_theta"] * magnetic_field["B^theta_z"]
        + magnetic_field["B^zeta"] * magnetic_field["B_zeta_z"]
        + magnetic_field["B_zeta"] * magnetic_field["B^zeta_z"]
    )

    # magnetic pressure gradient
    magnetic_pressure["grad(|B|^2)"] = (
        magnetic_pressure["grad(|B|^2)_rho"] * con_basis["e^rho"]
        + magnetic_pressure["grad(|B|^2)_theta"] * con_basis["e^theta"]
        + magnetic_pressure["grad(|B|^2)_zeta"] * con_basis["e^zeta"]
    )

    # magnetic pressure gradient magnitude
    magnetic_pressure["|grad(|B|^2)|"] = jnp.sqrt(
        magnetic_pressure["grad(|B|^2)_rho"] ** 2
        * dot(con_basis["e^rho"], con_basis["e^rho"], 0)
        + magnetic_pressure["grad(|B|^2)_theta"] ** 2
        * dot(con_basis["e^theta"], con_basis["e^theta"], 0)
        + magnetic_pressure["grad(|B|^2)_zeta"] ** 2
        * dot(con_basis["e^zeta"], con_basis["e^zeta"], 0)
    )

    # scaled magnetic pressure
    magnetic_pressure["Bpressure"] = magnetic_pressure["|grad(|B|^2)|"] / (2 * mu_0)

    return (
        magnetic_pressure,
        current_density,
        magnetic_field,
        con_basis,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    )


def compute_magnetic_tension(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute magnetic tension vector and its magnitude.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    magnetic_tension : dict
        dictionary of ndarray, shape(num_nodes,), of magnetic tension vector.
        Keys are of the form 'grad(B)' for the vector form and '|grad(B)|' for its
        magnitude.
    current_density : dict
        dictionary of ndarray, shape(num_nodes,), of current density components.
        Keys are of the form 'J^x_y' meaning the contravariant (J^x)
        component of the current, with the derivative wrt to y.
    magnetic_field: dict
        dictionary of ndarray, shape(num_nodes,) of magnetic field components.
        Keys are of the form 'B_x_y' or 'B^x_y', meaning the covariant (B_x)
        or contravariant (B^x) component of the magnetic field, with the
        derivative wrt to y.
    con_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of contravariant basis vectors.
        Keys are of the form 'e^x_y', meaning the contravariant basis vector
        in the x direction, differentiated wrt to y.
    jacobian : dict
        dictionary of ndarray, shape(num_nodes,), of coordinate system jacobian.
        Keys are of the form 'g_x' meaning the x derivative of the coordinate
        system jacobian g.
    cov_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of covariant basis vectors.
        Keys are of the form 'e_x_y', meaning the covariant basis vector in
        the x direction, differentiated wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.
    profiles : dict
        dictionary of ndarray, shape(num_nodes,) of profiles.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    (
        current_density,
        magnetic_field,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    ) = compute_current_density(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    # contravariant basis vectors
    con_basis = {}
    con_basis["e^rho"] = (
        cross(cov_basis["e_theta"], cov_basis["e_zeta"], 0) / jacobian["g"]
    )
    con_basis["e^theta"] = (
        cross(cov_basis["e_zeta"], cov_basis["e_rho"], 0) / jacobian["g"]
    )
    con_basis["e^zeta"] = jnp.array(
        [toroidal_coords["0"], 1 / toroidal_coords["R"], toroidal_coords["0"]]
    )

    magnetic_tension = {}

    # magnetic tension contravariant vector
    magnetic_tension["(B*grad(|B|))B"] = (
        (
            magnetic_field["B^theta"] * magnetic_field["B^theta_t"]
            + magnetic_field["B^zeta"] * magnetic_field["B^theta_z"]
        )
        * cov_basis["e_theta"]
        + (
            magnetic_field["B^theta"] * magnetic_field["B^zeta_t"]
            + magnetic_field["B^zeta"] * magnetic_field["B^zeta_z"]
        )
        * cov_basis["e_zeta"]
        + magnetic_field["B^theta"] ** 2 * cov_basis["e_theta_t"]
        + magnetic_field["B^zeta"] ** 2 * cov_basis["e_zeta_z"]
        + magnetic_field["B^theta"]
        * magnetic_field["B^zeta"]
        * (cov_basis["e_theta_z"] + cov_basis["e_zeta_t"])
    )

    # magnetic tension contravariant components
    magnetic_tension["((B*grad(|B|))B)^rho"] = dot(
        magnetic_tension["(B*grad(|B|))B"], con_basis["e^rho"], 0
    )
    magnetic_tension["((B*grad(|B|))B)^theta"] = dot(
        magnetic_tension["(B*grad(|B|))B"], con_basis["e^theta"], 0
    )
    magnetic_tension["((B*grad(|B|))B)^zeta"] = dot(
        magnetic_tension["(B*grad(|B|))B"], con_basis["e^zeta"], 0
    )

    # magnetic tension magnitude
    magnetic_tension["|(B*grad(|B|))B|"] = jnp.sqrt(
        magnetic_tension["((B*grad(|B|))B)^rho"] ** 2
        * dot(cov_basis["e_rho"], cov_basis["e_rho"], 0)
        + magnetic_tension["((B*grad(|B|))B)^theta"] ** 2
        * dot(cov_basis["e_theta"], cov_basis["e_theta"], 0)
        + magnetic_tension["((B*grad(|B|))B)^zeta"] ** 2
        * dot(cov_basis["e_zeta"], cov_basis["e_zeta"], 0)
    )

    # scaled magnetic tension
    magnetic_tension["Btension"] = magnetic_tension["|(B*grad(|B|))B|"] / (2 * mu_0)

    return (
        magnetic_tension,
        current_density,
        magnetic_field,
        con_basis,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    )


def compute_force_error(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute force error components.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    force_error : dict
        dictionary of ndarray, shape(num_nodes,), of force error components.
        Keys are of the form 'F_x' meaning the covariant (F_x) component of the
        force error.
    current_density : dict
        dictionary of ndarray, shape(num_nodes,), of current density components.
        Keys are of the form 'J^x_y' meaning the contravariant (J^x)
        component of the current, with the derivative wrt to y.
    magnetic_field: dict
        dictionary of ndarray, shape(num_nodes,) of magnetic field components.
        Keys are of the form 'B_x_y' or 'B^x_y', meaning the covariant (B_x)
        or contravariant (B^x) component of the magnetic field, with the
        derivative wrt to y.
    jacobian : dict
        dictionary of ndarray, shape(num_nodes,), of coordinate system jacobian.
        Keys are of the form 'g_x' meaning the x derivative of the coordinate
        system jacobian g.
    cov_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of covariant basis vectors.
        Keys are of the form 'e_x_y', meaning the covariant basis vector in
        the x direction, differentiated wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.
    profiles : dict
        dictionary of ndarray, shape(num_nodes,) of profiles.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    (
        current_density,
        magnetic_field,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    ) = compute_current_density(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    force_error = {}
    force_error["F_rho"] = -profiles["p_r"] + jacobian["g"] * (
        current_density["J^theta"] * magnetic_field["B^zeta"]
        - current_density["J^zeta"] * magnetic_field["B^theta"]
    )
    force_error["F_theta"] = (
        jacobian["g"] * current_density["J^rho"] * magnetic_field["B^zeta"]
    )
    force_error["F_zeta"] = -profiles["iota"] * force_error["F_theta"]
    force_error["F_beta"] = force_error["F_theta"]

    return (
        force_error,
        current_density,
        magnetic_field,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    )


def compute_force_error_magnitude(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute force error magnitude.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    force_error : dict
        dictionary of ndarray, shape(num_nodes,), of force error components.
        Keys are of the form 'F_x' meaning the covariant (F_x) component of the
        force error.
    current_density : dict
        dictionary of ndarray, shape(num_nodes,), of current density components.
        Keys are of the form 'J^x_y' meaning the contravariant (J^x)
        component of the current, with the derivative wrt to y.
    magnetic_field: dict
        dictionary of ndarray, shape(num_nodes,) of magnetic field components.
        Keys are of the form 'B_x_y' or 'B^x_y', meaning the covariant (B_x)
        or contravariant (B^x) component of the magnetic field, with the
        derivative wrt to y.
    con_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of contravariant basis vectors.
        Keys are of the form 'e^x_y', meaning the contravariant basis vector
        in the x direction, differentiated wrt to y.
    jacobian : dict
        dictionary of ndarray, shape(num_nodes,), of coordinate system jacobian.
        Keys are of the form 'g_x' meaning the x derivative of the coordinate
        system jacobian g.
    cov_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of covariant basis vectors.
        Keys are of the form 'e_x_y', meaning the covariant basis vector in
        the x direction, differentiated wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.
    profiles : dict
        dictionary of ndarray, shape(num_nodes,) of profiles.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    (
        force_error,
        current_density,
        magnetic_field,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    ) = compute_force_error(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    # contravariant basis vectors
    con_basis = {}
    con_basis["e^rho"] = (
        cross(cov_basis["e_theta"], cov_basis["e_zeta"], 0) / jacobian["g"]
    )
    con_basis["e^theta"] = (
        cross(cov_basis["e_zeta"], cov_basis["e_rho"], 0) / jacobian["g"]
    )
    con_basis["e^zeta"] = jnp.array(
        [toroidal_coords["0"], 1 / toroidal_coords["R"], toroidal_coords["0"]]
    )

    force_error["F"] = (
        force_error["F_rho"] * con_basis["e^rho"]
        + force_error["F_theta"] * con_basis["e^theta"]
        + force_error["F_zeta"] * con_basis["e^zeta"]
    )
    force_error["beta"] = con_basis["e^theta"] - profiles["iota"] * con_basis["e^zeta"]

    force_error["|grad(rho)|"] = jnp.sqrt(
        dot(con_basis["e^rho"], con_basis["e^rho"], 0)
    )
    force_error["|grad(p)|"] = jnp.sqrt(
        profiles["p_r"] ** 2 * force_error["|grad(rho)|"] ** 2
    )
    force_error["|beta|"] = jnp.sqrt(
        dot(con_basis["e^theta"], con_basis["e^theta"], 0)
        + profiles["iota"] ** 2 * dot(con_basis["e^zeta"], con_basis["e^zeta"], 0)
        - 2 * profiles["iota"] * dot(con_basis["e^theta"], con_basis["e^zeta"], 0)
    )

    force_error["|F|"] = jnp.sqrt(
        force_error["F_rho"] ** 2 * force_error["|grad(rho)|"] ** 2
        + force_error["F_beta"] ** 2 * force_error["|beta|"] ** 2
    )

    return (
        force_error,
        current_density,
        magnetic_field,
        con_basis,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    )


def compute_energy(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute MHD energy by quadrature sum.

    **REQUIRES 'quad' grid for correct results**

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    energy: float
        The scalar value of the integral of (B^2/(2*mu0) - p) over the plasma volume.
    magnetic_field: dict
        dictionary of ndarray, shape(num_nodes,) of magnetic field components.
        Keys are of the form 'B_x_y' or 'B^x_y', meaning the covariant (B_x)
        or contravariant (B^x) component of the magnetic field, with the
        derivative wrt to y.
    jacobian : dict
        dictionary of ndarray, shape(num_nodes,), of coordinate system jacobian.
        Keys are of the form 'g_x' meaning the x derivative of the coordinate
        system jacobian g.
    cov_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of covariant basis vectors.
        Keys are of the form 'e_x_y', meaning the covariant basis vector in
        the x direction, differentiated wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.
    profiles : dict
        dictionary of ndarray, shape(num_nodes,) of profiles.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    (
        magnetic_field,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    ) = compute_magnetic_field_magnitude(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    NFP = R_transform.grid.NFP
    weights = R_transform.grid.weights

    energy = {}
    energy["W_p"] = -jnp.sum(profiles["p"] * jnp.abs(jacobian["g"]) * weights) * NFP
    energy["W_B"] = (
        jnp.sum(magnetic_field["|B|"] ** 2 * jnp.abs(jacobian["g"]) * weights)
        / (2 * mu_0)
        * NFP
    )
    energy["W"] = energy["W_B"] - energy["W_p"]
    energy["beta"] = jnp.abs(energy["W_p"] / energy["W_B"])

    return (energy, magnetic_field, jacobian, cov_basis, toroidal_coords, profiles)


def compute_quasisymmetry(
    Psi,
    R_lmn,
    Z_lmn,
    L_lmn,
    p_l,
    i_l,
    R_transform,
    Z_transform,
    L_transform,
    p_transform,
    i_transform,
    zeta_ratio=1.0,
):
    """Compute quasisymmetry metrics.

    Parameters
    ----------
    Psi : float
        total toroidal flux (in Webers) within the last closed flux surface
    R_lmn : ndarray
        spectral coefficients of R(rho,theta,zeta) -- flux surface R coordinate
    Z_lmn : ndarray
        spectral coefficients of Z(rho,theta,zeta) -- flux surface Z coordiante
    L_lmn : ndarray
        spectral coefficients of lambda(rho,theta,zeta) -- sfl coordinate map
    p_l : ndarray
        spectral coefficients of p(rho) -- pressure profile
    i_l : ndarray
        spectral coefficients of iota(rho) -- rotational transform profile
    R_transform : Transform
        transforms R_lmn coefficients to real space
    Z_transform : Transform
        transforms Z_lmn coefficients to real space
    L_transform : Transform
        transforms L_lmn coefficients to real space
    p_transform : Transform
        transforms p_l coefficients to real space
    i_transform : Transform
        transforms i_l coefficients to real space
    zeta_ratio : float
        scale factor for zeta derivatives. Setting to zero effectively solves
        for individual tokamak solutions at each toroidal plane,
        setting to 1 solves for a stellarator. (Default value = 1.0)

    Returns
    -------
    quasisymmetry: dict
        dictionary of ndarray, shape(num_nodes,), of quasisymmetry components.
        The triple product metric has the key 'QS_TP',
        and the flux function metric has the key 'QS_FF'.
    current_density : dict
        dictionary of ndarray, shape(num_nodes,), of current density components.
        Keys are of the form 'J^x_y' meaning the contravariant (J^x)
        component of the current, with the derivative wrt to y.
    magnetic_field: dict
        dictionary of ndarray, shape(num_nodes,) of magnetic field components.
        Keys are of the form 'B_x_y' or 'B^x_y', meaning the covariant (B_x)
        or contravariant (B^x) component of the magnetic field, with the
        derivative wrt to y.
    jacobian : dict
        dictionary of ndarray, shape(num_nodes,), of coordinate system jacobian.
        Keys are of the form 'g_x' meaning the x derivative of the coordinate
        system jacobian g.
    cov_basis : dict
        dictionary of ndarray, shape(3,num_nodes), of covariant basis vectors.
        Keys are of the form 'e_x_y', meaning the covariant basis vector in
        the x direction, differentiated wrt to y.
    toroidal_coords : dict
        dictionary of ndarray, shape(num_nodes,) of toroidal coordinates.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.
    profiles : dict
        dictionary of ndarray, shape(num_nodes,) of profiles.
        Keys are of the form 'X_y' meaning the derivative of X wrt to y.

    """
    # prerequisites
    (
        current_density,
        magnetic_field,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    ) = compute_current_density(
        Psi,
        R_lmn,
        Z_lmn,
        L_lmn,
        p_l,
        i_l,
        R_transform,
        Z_transform,
        L_transform,
        p_transform,
        i_transform,
        zeta_ratio,
    )

    axis = i_transform.grid.axis

    # toroidal coordinate 2nd derivatives
    toroidal_coords["R_rtt"] = R_transform.transform(R_lmn, 1, 2, 0)
    toroidal_coords["Z_rtt"] = Z_transform.transform(Z_lmn, 1, 2, 0)
    toroidal_coords["R_rtz"] = R_transform.transform(R_lmn, 1, 1, 1) * zeta_ratio
    toroidal_coords["Z_rtz"] = Z_transform.transform(Z_lmn, 1, 1, 1) * zeta_ratio
    toroidal_coords["R_rzz"] = R_transform.transform(R_lmn, 1, 0, 2) * zeta_ratio
    toroidal_coords["Z_rzz"] = Z_transform.transform(Z_lmn, 1, 0, 2) * zeta_ratio
    toroidal_coords["R_ttt"] = R_transform.transform(R_lmn, 0, 3, 0)
    toroidal_coords["Z_ttt"] = Z_transform.transform(Z_lmn, 0, 3, 0)
    toroidal_coords["R_ttz"] = R_transform.transform(R_lmn, 0, 2, 1) * zeta_ratio
    toroidal_coords["Z_ttz"] = Z_transform.transform(Z_lmn, 0, 2, 1) * zeta_ratio
    toroidal_coords["R_tzz"] = R_transform.transform(R_lmn, 0, 1, 2) * zeta_ratio
    toroidal_coords["Z_tzz"] = Z_transform.transform(Z_lmn, 0, 1, 2) * zeta_ratio
    toroidal_coords["R_zzz"] = R_transform.transform(R_lmn, 0, 0, 3) * zeta_ratio
    toroidal_coords["Z_zzz"] = Z_transform.transform(Z_lmn, 0, 0, 3) * zeta_ratio

    # lambda derivatives
    toroidal_coords["lambda_ttt"] = L_transform.transform(L_lmn, 0, 3, 0)
    toroidal_coords["lambda_ttz"] = L_transform.transform(L_lmn, 0, 2, 1) * zeta_ratio
    toroidal_coords["lambda_tzz"] = L_transform.transform(L_lmn, 0, 1, 2) * zeta_ratio
    toroidal_coords["lambda_zzz"] = L_transform.transform(L_lmn, 0, 0, 3) * zeta_ratio

    # covariant basis derivatives
    cov_basis["e_rho_tt"] = jnp.array(
        [toroidal_coords["R_rtt"], toroidal_coords["0"], toroidal_coords["Z_rtt"]]
    )
    cov_basis["e_theta_tt"] = jnp.array(
        [toroidal_coords["R_ttt"], toroidal_coords["0"], toroidal_coords["Z_ttt"]]
    )
    cov_basis["e_zeta_tt"] = jnp.array(
        [toroidal_coords["R_ttz"], toroidal_coords["R_tt"], toroidal_coords["Z_ttz"]]
    )
    cov_basis["e_rho_zz"] = jnp.array(
        [toroidal_coords["R_rzz"], toroidal_coords["0"], toroidal_coords["Z_rzz"]]
    )
    cov_basis["e_theta_zz"] = jnp.array(
        [toroidal_coords["R_tzz"], toroidal_coords["0"], toroidal_coords["Z_tzz"]]
    )
    cov_basis["e_zeta_zz"] = jnp.array(
        [toroidal_coords["R_zzz"], toroidal_coords["R_zz"], toroidal_coords["Z_zzz"]]
    )
    cov_basis["e_rho_tz"] = jnp.array(
        [toroidal_coords["R_rtz"], toroidal_coords["0"], toroidal_coords["Z_rtz"]]
    )
    cov_basis["e_theta_tz"] = jnp.array(
        [toroidal_coords["R_ttz"], toroidal_coords["0"], toroidal_coords["Z_ttz"]]
    )
    cov_basis["e_zeta_tz"] = jnp.array(
        [toroidal_coords["R_tzz"], toroidal_coords["R_tz"], toroidal_coords["Z_tzz"]]
    )

    # jacobian derivatives
    jacobian["g_tt"] = (
        dot(
            cov_basis["e_rho_tt"],
            cross(cov_basis["e_theta"], cov_basis["e_zeta"], 0),
            0,
        )
        + dot(
            cov_basis["e_rho"],
            cross(cov_basis["e_theta_tt"], cov_basis["e_zeta"], 0),
            0,
        )
        + dot(
            cov_basis["e_rho"],
            cross(cov_basis["e_theta"], cov_basis["e_zeta_tt"], 0),
            0,
        )
        + 2
        * dot(
            cov_basis["e_rho_t"],
            cross(cov_basis["e_theta_t"], cov_basis["e_zeta"], 0),
            0,
        )
        + 2
        * dot(
            cov_basis["e_rho_t"],
            cross(cov_basis["e_theta"], cov_basis["e_zeta_t"], 0),
            0,
        )
        + 2
        * dot(
            cov_basis["e_rho"],
            cross(cov_basis["e_theta_t"], cov_basis["e_zeta_t"], 0),
            0,
        )
    )
    jacobian["g_zz"] = (
        dot(
            cov_basis["e_rho_zz"],
            cross(cov_basis["e_theta"], cov_basis["e_zeta"], 0),
            0,
        )
        + dot(
            cov_basis["e_rho"],
            cross(cov_basis["e_theta_zz"], cov_basis["e_zeta"], 0),
            0,
        )
        + dot(
            cov_basis["e_rho"],
            cross(cov_basis["e_theta"], cov_basis["e_zeta_zz"], 0),
            0,
        )
        + 2
        * dot(
            cov_basis["e_rho_z"],
            cross(cov_basis["e_theta_z"], cov_basis["e_zeta"], 0),
            0,
        )
        + 2
        * dot(
            cov_basis["e_rho_z"],
            cross(cov_basis["e_theta"], cov_basis["e_zeta_z"], 0),
            0,
        )
        + 2
        * dot(
            cov_basis["e_rho"],
            cross(cov_basis["e_theta_z"], cov_basis["e_zeta_z"], 0),
            0,
        )
    )
    jacobian["g_tz"] = (
        dot(
            cov_basis["e_rho_tz"],
            cross(cov_basis["e_theta"], cov_basis["e_zeta"], 0),
            0,
        )
        + dot(
            cov_basis["e_rho_z"],
            cross(cov_basis["e_theta_t"], cov_basis["e_zeta"], 0),
            0,
        )
        + dot(
            cov_basis["e_rho_z"],
            cross(cov_basis["e_theta"], cov_basis["e_zeta_t"], 0),
            0,
        )
        + dot(
            cov_basis["e_rho_t"],
            cross(cov_basis["e_theta_z"], cov_basis["e_zeta"], 0),
            0,
        )
        + dot(
            cov_basis["e_rho"],
            cross(cov_basis["e_theta_tz"], cov_basis["e_zeta"], 0),
            0,
        )
        + dot(
            cov_basis["e_rho"],
            cross(cov_basis["e_theta_z"], cov_basis["e_zeta_t"], 0),
            0,
        )
        + dot(
            cov_basis["e_rho_t"],
            cross(cov_basis["e_theta"], cov_basis["e_zeta_z"], 0),
            0,
        )
        + dot(
            cov_basis["e_rho"],
            cross(cov_basis["e_theta_t"], cov_basis["e_zeta_z"], 0),
            0,
        )
        + dot(
            cov_basis["e_rho"],
            cross(cov_basis["e_theta"], cov_basis["e_zeta_tz"], 0),
            0,
        )
    )

    # B covariant component derivatives
    magnetic_field["B_theta_t"] = dot(
        magnetic_field["B_t"], cov_basis["e_theta"], 0
    ) + dot(magnetic_field["B"], cov_basis["e_theta_t"], 0)
    magnetic_field["B_zeta_z"] = dot(
        magnetic_field["B_z"], cov_basis["e_zeta"], 0
    ) + dot(magnetic_field["B"], cov_basis["e_zeta_z"], 0)

    # B contravariant component derivatives
    magnetic_field["B0_tt"] = -(
        profiles["psi_r"]
        / (2 * jnp.pi * jacobian["g"] ** 2)
        * (jacobian["g_tt"] - 2 * jacobian["g_t"] ** 2 / jacobian["g"])
    )
    magnetic_field["B0_zz"] = -(
        profiles["psi_r"]
        / (2 * jnp.pi * jacobian["g"] ** 2)
        * (jacobian["g_zz"] - 2 * jacobian["g_z"] ** 2 / jacobian["g"])
    )
    magnetic_field["B0_tz"] = -(
        profiles["psi_r"]
        / (2 * jnp.pi * jacobian["g"] ** 2)
        * (jacobian["g_tz"] - 2 * jacobian["g_t"] * jacobian["g_z"] / jacobian["g"])
    )
    magnetic_field["B^theta_tt"] = magnetic_field["B0_tt"] * (
        profiles["iota"] - toroidal_coords["lambda_z"]
    )
    -2 * magnetic_field["B0_t"] * toroidal_coords["lambda_tz"]
    -magnetic_field["B0"] * toroidal_coords["lambda_ttz"]
    magnetic_field["B^theta_zz"] = magnetic_field["B0_zz"] * (
        profiles["iota"] - toroidal_coords["lambda_z"]
    )
    -2 * magnetic_field["B0_z"] * toroidal_coords["lambda_zz"]
    -magnetic_field["B0"] * toroidal_coords["lambda_zzz"]
    magnetic_field["B^theta_tz"] = magnetic_field["B0_tz"] * (
        profiles["iota"] - toroidal_coords["lambda_z"]
    )
    -magnetic_field["B0_t"] * toroidal_coords["lambda_zz"]
    -magnetic_field["B0_z"] * toroidal_coords["lambda_tz"]
    -magnetic_field["B0"] * toroidal_coords["lambda_tzz"]
    magnetic_field["B^zeta_tt"] = magnetic_field["B0_tt"] * (
        1 + toroidal_coords["lambda_t"]
    )
    +2 * magnetic_field["B0_t"] * toroidal_coords["lambda_tt"]
    +magnetic_field["B0"] * toroidal_coords["lambda_ttt"]
    magnetic_field["B^zeta_zz"] = magnetic_field["B0_zz"] * (
        1 + toroidal_coords["lambda_t"]
    )
    +2 * magnetic_field["B0_z"] * toroidal_coords["lambda_tz"]
    +magnetic_field["B0"] * toroidal_coords["lambda_tzz"]
    magnetic_field["B^zeta_tz"] = magnetic_field["B0_tz"] * (
        1 + toroidal_coords["lambda_t"]
    )
    +magnetic_field["B0_t"] * toroidal_coords["lambda_tz"]
    +magnetic_field["B0_z"] * toroidal_coords["lambda_tt"]
    +magnetic_field["B0"] * toroidal_coords["lambda_ttz"]

    # |B|
    magnetic_field["|B|"] = jnp.sqrt(
        magnetic_field["B^theta"] ** 2
        * dot(cov_basis["e_theta"], cov_basis["e_theta"], 0)
        + magnetic_field["B^zeta"] ** 2
        * dot(cov_basis["e_zeta"], cov_basis["e_zeta"], 0)
        + 2
        * magnetic_field["B^theta"]
        * magnetic_field["B^zeta"]
        * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
    )

    # |B| derivatives
    magnetic_field["|B|_t"] = (
        magnetic_field["B^theta"]
        * (
            magnetic_field["B^zeta_t"]
            * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^theta_t"]
            * dot(cov_basis["e_theta"], cov_basis["e_theta"], 0)
            + magnetic_field["B^theta"]
            * dot(cov_basis["e_theta_t"], cov_basis["e_theta"], 0)
        )
        + magnetic_field["B^zeta"]
        * (
            magnetic_field["B^theta_t"]
            * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^zeta_t"]
            * dot(cov_basis["e_zeta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^zeta"]
            * dot(cov_basis["e_zeta_t"], cov_basis["e_zeta"], 0)
        )
        + magnetic_field["B^theta"]
        * magnetic_field["B^zeta"]
        * (
            dot(cov_basis["e_theta_t"], cov_basis["e_zeta"], 0)
            + dot(cov_basis["e_zeta_t"], cov_basis["e_theta"], 0)
        )
    ) / magnetic_field["|B|"]
    magnetic_field["|B|_z"] = (
        magnetic_field["B^theta"]
        * (
            magnetic_field["B^zeta_z"]
            * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^theta_z"]
            * dot(cov_basis["e_theta"], cov_basis["e_theta"], 0)
            + magnetic_field["B^theta"]
            * dot(cov_basis["e_theta_z"], cov_basis["e_theta"], 0)
        )
        + magnetic_field["B^zeta"]
        * (
            magnetic_field["B^theta_z"]
            * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^zeta_z"]
            * dot(cov_basis["e_zeta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^zeta"]
            * dot(cov_basis["e_zeta_z"], cov_basis["e_zeta"], 0)
        )
        + magnetic_field["B^theta"]
        * magnetic_field["B^zeta"]
        * (
            dot(cov_basis["e_theta_z"], cov_basis["e_zeta"], 0)
            + dot(cov_basis["e_zeta_z"], cov_basis["e_theta"], 0)
        )
    ) / magnetic_field["|B|"]
    magnetic_field["|B|_tt"] = (
        magnetic_field["B^theta_t"]
        * (
            magnetic_field["B^zeta_t"]
            * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^theta_t"]
            * dot(cov_basis["e_theta"], cov_basis["e_theta"], 0)
            + magnetic_field["B^theta"]
            * dot(cov_basis["e_theta_t"], cov_basis["e_theta"], 0)
        )
        + magnetic_field["B^theta"]
        * (
            magnetic_field["B^zeta_tt"]
            * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^theta_tt"]
            * dot(cov_basis["e_theta"], cov_basis["e_theta"], 0)
            + magnetic_field["B^theta_t"]
            * dot(cov_basis["e_theta_t"], cov_basis["e_theta"], 0)
        )
        + magnetic_field["B^theta"]
        * (
            magnetic_field["B^zeta_t"]
            * (
                dot(cov_basis["e_theta_t"], cov_basis["e_zeta"], 0)
                + dot(cov_basis["e_theta"], cov_basis["e_zeta_t"], 0)
            )
            + 2
            * magnetic_field["B^theta_t"]
            * dot(cov_basis["e_theta_t"], cov_basis["e_theta"], 0)
            + magnetic_field["B^theta"]
            * (
                dot(cov_basis["e_theta_tt"], cov_basis["e_theta"], 0)
                + dot(cov_basis["e_theta_t"], cov_basis["e_theta_t"], 0)
            )
        )
        + magnetic_field["B^zeta_t"]
        * (
            magnetic_field["B^theta_t"]
            * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^zeta_t"]
            * dot(cov_basis["e_zeta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^zeta"]
            * dot(cov_basis["e_zeta_t"], cov_basis["e_zeta"], 0)
        )
        + magnetic_field["B^zeta"]
        * (
            magnetic_field["B^theta_tt"]
            * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^zeta_tt"]
            * dot(cov_basis["e_zeta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^zeta_t"]
            * dot(cov_basis["e_zeta_t"], cov_basis["e_zeta"], 0)
        )
        + magnetic_field["B^zeta"]
        * (
            magnetic_field["B^theta_t"]
            * (
                dot(cov_basis["e_theta_t"], cov_basis["e_zeta"], 0)
                + dot(cov_basis["e_theta"], cov_basis["e_zeta_t"], 0)
            )
            + 2
            * magnetic_field["B^zeta_t"]
            * dot(cov_basis["e_zeta_t"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^zeta"]
            * (
                dot(cov_basis["e_zeta_tt"], cov_basis["e_zeta"], 0)
                + dot(cov_basis["e_zeta_t"], cov_basis["e_zeta_t"], 0)
            )
        )
        + (
            magnetic_field["B^theta_t"] * magnetic_field["B^zeta"]
            + magnetic_field["B^theta"] * magnetic_field["B^zeta_t"]
        )
        * (
            dot(cov_basis["e_theta_t"], cov_basis["e_zeta"], 0)
            + dot(cov_basis["e_zeta_t"], cov_basis["e_theta"], 0)
        )
        + magnetic_field["B^theta"]
        * magnetic_field["B^zeta"]
        * (
            dot(cov_basis["e_theta_tt"], cov_basis["e_zeta"], 0)
            + dot(cov_basis["e_zeta_tt"], cov_basis["e_theta"], 0)
            + 2 * dot(cov_basis["e_zeta_t"], cov_basis["e_theta_t"], 0)
        )
    ) / magnetic_field["|B|"] - magnetic_field["|B|_t"] ** 2 / magnetic_field["|B|"]
    magnetic_field["|B|_zz"] = (
        magnetic_field["B^theta_z"]
        * (
            magnetic_field["B^zeta_z"]
            * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^theta_z"]
            * dot(cov_basis["e_theta"], cov_basis["e_theta"], 0)
            + magnetic_field["B^theta"]
            * dot(cov_basis["e_theta_z"], cov_basis["e_theta"], 0)
        )
        + magnetic_field["B^theta"]
        * (
            magnetic_field["B^zeta_zz"]
            * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^theta_zz"]
            * dot(cov_basis["e_theta"], cov_basis["e_theta"], 0)
            + magnetic_field["B^theta_z"]
            * dot(cov_basis["e_theta_z"], cov_basis["e_theta"], 0)
        )
        + magnetic_field["B^theta"]
        * (
            magnetic_field["B^zeta_z"]
            * (
                dot(cov_basis["e_theta_z"], cov_basis["e_zeta"], 0)
                + dot(cov_basis["e_theta"], cov_basis["e_zeta_z"], 0)
            )
            + 2
            * magnetic_field["B^theta_z"]
            * dot(cov_basis["e_theta_z"], cov_basis["e_theta"], 0)
            + magnetic_field["B^theta"]
            * (
                dot(cov_basis["e_theta_zz"], cov_basis["e_theta"], 0)
                + dot(cov_basis["e_theta_z"], cov_basis["e_theta_z"], 0)
            )
        )
        + magnetic_field["B^zeta_z"]
        * (
            magnetic_field["B^theta_z"]
            * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^zeta_z"]
            * dot(cov_basis["e_zeta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^zeta"]
            * dot(cov_basis["e_zeta_z"], cov_basis["e_zeta"], 0)
        )
        + magnetic_field["B^zeta"]
        * (
            magnetic_field["B^theta_zz"]
            * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^zeta_zz"]
            * dot(cov_basis["e_zeta"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^zeta_z"]
            * dot(cov_basis["e_zeta_z"], cov_basis["e_zeta"], 0)
        )
        + magnetic_field["B^zeta"]
        * (
            magnetic_field["B^theta_z"]
            * (
                dot(cov_basis["e_theta_z"], cov_basis["e_zeta"], 0)
                + dot(cov_basis["e_theta"], cov_basis["e_zeta_z"], 0)
            )
            + 2
            * magnetic_field["B^zeta_z"]
            * dot(cov_basis["e_zeta_z"], cov_basis["e_zeta"], 0)
            + magnetic_field["B^zeta"]
            * (
                dot(cov_basis["e_zeta_zz"], cov_basis["e_zeta"], 0)
                + dot(cov_basis["e_zeta_z"], cov_basis["e_zeta_z"], 0)
            )
        )
        + (
            magnetic_field["B^theta_z"] * magnetic_field["B^zeta"]
            + magnetic_field["B^theta"] * magnetic_field["B^zeta_z"]
        )
        * (
            dot(cov_basis["e_theta_z"], cov_basis["e_zeta"], 0)
            + dot(cov_basis["e_zeta_z"], cov_basis["e_theta"], 0)
        )
        + magnetic_field["B^theta"]
        * magnetic_field["B^zeta"]
        * (
            dot(cov_basis["e_theta_zz"], cov_basis["e_zeta"], 0)
            + dot(cov_basis["e_zeta_zz"], cov_basis["e_theta"], 0)
            + 2 * dot(cov_basis["e_theta_z"], cov_basis["e_zeta_z"], 0)
        )
    ) / magnetic_field["|B|"] - magnetic_field["|B|_z"] ** 2 / magnetic_field["|B|"]
    magnetic_field["|B|_tz"] = (
        (
            magnetic_field["B^theta_z"]
            * (
                magnetic_field["B^zeta_t"]
                * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
                + magnetic_field["B^theta_t"]
                * dot(cov_basis["e_theta"], cov_basis["e_theta"], 0)
                + magnetic_field["B^theta"]
                * dot(cov_basis["e_theta_t"], cov_basis["e_theta"], 0)
            )
            + magnetic_field["B^theta"]
            * (
                magnetic_field["B^zeta_tz"]
                * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
                + magnetic_field["B^theta_tz"]
                * dot(cov_basis["e_theta"], cov_basis["e_theta"], 0)
                + magnetic_field["B^theta_z"]
                * dot(cov_basis["e_theta_t"], cov_basis["e_theta"], 0)
            )
            + magnetic_field["B^theta"]
            * (
                magnetic_field["B^zeta_t"]
                * (
                    dot(cov_basis["e_theta_z"], cov_basis["e_zeta"], 0)
                    + dot(cov_basis["e_theta"], cov_basis["e_zeta_z"], 0)
                )
                + 2
                * magnetic_field["B^theta_t"]
                * dot(cov_basis["e_theta_z"], cov_basis["e_theta"], 0)
                + magnetic_field["B^theta"]
                * (
                    dot(cov_basis["e_theta_tz"], cov_basis["e_theta"], 0)
                    + dot(cov_basis["e_theta_t"], cov_basis["e_theta_z"], 0)
                )
            )
            + magnetic_field["B^zeta_z"]
            * (
                magnetic_field["B^theta_t"]
                * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
                + magnetic_field["B^zeta_t"]
                * dot(cov_basis["e_zeta"], cov_basis["e_zeta"], 0)
                + magnetic_field["B^zeta"]
                * dot(cov_basis["e_zeta_t"], cov_basis["e_zeta"], 0)
            )
            + magnetic_field["B^zeta"]
            * (
                magnetic_field["B^theta_tz"]
                * dot(cov_basis["e_theta"], cov_basis["e_zeta"], 0)
                + magnetic_field["B^zeta_tz"]
                * dot(cov_basis["e_zeta"], cov_basis["e_zeta"], 0)
                + magnetic_field["B^zeta_z"]
                * dot(cov_basis["e_zeta_t"], cov_basis["e_zeta"], 0)
            )
            + magnetic_field["B^zeta"]
            * (
                magnetic_field["B^theta_t"]
                * (
                    dot(cov_basis["e_theta_z"], cov_basis["e_zeta"], 0)
                    + dot(cov_basis["e_theta"], cov_basis["e_zeta_z"], 0)
                )
                + 2
                * magnetic_field["B^zeta_t"]
                * dot(cov_basis["e_zeta_z"], cov_basis["e_zeta"], 0)
                + magnetic_field["B^zeta"]
                * (
                    dot(cov_basis["e_zeta_tz"], cov_basis["e_zeta"], 0)
                    + dot(cov_basis["e_zeta_t"], cov_basis["e_zeta_z"], 0)
                )
            )
            + (
                magnetic_field["B^theta_z"] * magnetic_field["B^zeta"]
                + magnetic_field["B^theta"] * magnetic_field["B^zeta_z"]
            )
            * (
                dot(cov_basis["e_theta_t"], cov_basis["e_zeta"], 0)
                + dot(cov_basis["e_zeta_t"], cov_basis["e_theta"], 0)
            )
            + magnetic_field["B^theta"]
            * magnetic_field["B^zeta"]
            * (
                dot(cov_basis["e_theta_tz"], cov_basis["e_zeta"], 0)
                + dot(cov_basis["e_zeta_tz"], cov_basis["e_theta"], 0)
                + dot(cov_basis["e_theta_t"], cov_basis["e_zeta_z"], 0)
                + dot(cov_basis["e_zeta_t"], cov_basis["e_theta_z"], 0)
            )
        )
        / magnetic_field["|B|"]
        - magnetic_field["|B|_t"] * magnetic_field["|B|_z"] / magnetic_field["|B|"]
    )

    # contravariant basis vectors
    con_basis = {}
    con_basis["e^rho"] = (
        cross(cov_basis["e_theta"], cov_basis["e_zeta"], 0) / jacobian["g"]
    )
    con_basis["e^theta"] = (
        cross(cov_basis["e_zeta"], cov_basis["e_rho"], 0) / jacobian["g"]
    )
    con_basis["e^zeta"] = jnp.array(
        [toroidal_coords["0"], 1 / toroidal_coords["R"], toroidal_coords["0"]]
    )

    quasisymmetry = {}

    quasisymmetry["|grad(rho)|"] = jnp.sqrt(
        dot(con_basis["e^rho"], con_basis["e^rho"], 0)
    )
    quasisymmetry["|grad(psi)|"] = jnp.sqrt(
        profiles["psi_r"] ** 2 * quasisymmetry["|grad(rho)|"] ** 2
    )

    # B * grad(|B|) and derivatives
    quasisymmetry["B*grad(|B|)"] = (
        magnetic_field["B^theta"] * magnetic_field["|B|_t"]
        + magnetic_field["B^zeta"] * magnetic_field["|B|_z"]
    )
    quasisymmetry["B*grad(|B|)_t"] = (
        magnetic_field["B^theta_t"] * magnetic_field["|B|_t"]
        + magnetic_field["B^zeta_t"] * magnetic_field["|B|_z"]
        + magnetic_field["B^theta"] * magnetic_field["|B|_tt"]
        + magnetic_field["B^zeta"] * magnetic_field["|B|_tz"]
    )
    quasisymmetry["B*grad(|B|)_z"] = (
        magnetic_field["B^theta_z"] * magnetic_field["|B|_t"]
        + magnetic_field["B^zeta_z"] * magnetic_field["|B|_z"]
        + magnetic_field["B^theta"] * magnetic_field["|B|_tz"]
        + magnetic_field["B^zeta"] * magnetic_field["|B|_zz"]
    )

    # Triple Product QS metric (dimensionless)
    quasisymmetry["QS_TP"] = (
        magnetic_field["|B|_t"] * quasisymmetry["B*grad(|B|)_z"]
        - magnetic_field["|B|_z"] * quasisymmetry["B*grad(|B|)_t"]
    ) / (quasisymmetry["|grad(rho)|"] * magnetic_field["|B|"] ** 3)
    quasisymmetry["QS_TP"] = put(quasisymmetry["QS_TP"], axis, 0)

    # FIXME: singular points of QS Flux Function cause JAX issues
    # Singular points (where B*grad(|B|)=0) are not known a-priori.
    """
    # Flux Function QS metric (T*m)
    quasisymmetry["QS_FF"] = (
        profiles["psi_r"]
        / (jacobian["g"] * quasisymmetry["B*grad(|B|)"])
        * (
            +magnetic_field["B_zeta"] * magnetic_field["|B|_t"]
            - magnetic_field["B_theta"] * magnetic_field["|B|_z"]
        )
    )
    quasisymmetry["QS_FF_0"] = (
        profiles["psi_r"]
        * (
            magnetic_field["B_zeta_t"] * magnetic_field["|B|_t"]
            - magnetic_field["B_theta_t"] * magnetic_field["|B|_z"]
            + magnetic_field["B_zeta"] * magnetic_field["|B|_tt"]
            - magnetic_field["B_theta"] * magnetic_field["|B|_tz"]
        )
        / (
            jacobian["g_t"] * quasisymmetry["B*grad(|B|)"]
            + jacobian["g"] * quasisymmetry["B*grad(|B|)_t"]
        )
    )
    singular = jnp.where(quasisymmetry["B*grad(|B|)"] == 0)[0]
    quasisymmetry["QS_FF"] = put(
        quasisymmetry["QS_FF"], singular, quasisymmetry["QS_FF_0"][singular]
    )
    """

    return (
        quasisymmetry,
        current_density,
        magnetic_field,
        jacobian,
        cov_basis,
        toroidal_coords,
        profiles,
    )



import numpy as np
import warnings
import desc
import os
from termcolor import colored
from desc import set_device, config as desc_config

# we divide by zero in a few places but then overwrite with the
# correct values, so lets suppress annoying warnings about that
np.seterr(divide="ignore", invalid="ignore")

if os.environ.get("DESC_BACKEND") == "numpy":
    jnp = np
    use_jax = False
    set_device(kind="cpu")
    print(
        "DESC version {}, using numpy backend, version={}, dtype={}".format(
            desc.__version__, np.__version__, np.linspace(0, 1).dtype
        )
    )
else:
    if desc_config.get("device") is None:
        set_device("cpu")
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            import jax
            import jaxlib
            import jax.numpy as jnp
            from jax.config import config as jax_config

            jax_config.update("jax_enable_x64", True)
            x = jnp.linspace(0, 5)
            y = jnp.exp(x)
        use_jax = True
        print(
            "DESC version {}, using JAX backend, jax version={}, jaxlib version={}, dtype={}".format(
                desc.__version__, jax.__version__, jaxlib.__version__, y.dtype
            )
        )

    except:
        jnp = np
        x = jnp.linspace(0, 5)
        y = jnp.exp(x)
        use_jax = False
        set_device(kind="cpu")
        warnings.warn(colored("Failed to load JAX", "red"))
        print(
            "DESC version {}, using NumPy backend, version={}, dtype={}".format(
                desc.__version__, np.__version__, y.dtype
            )
        )
print(
    "Using device: {}, with {:.2f} GB available memory".format(
        desc_config.get("device"), desc_config.get("avail_mem")
    )
)

if use_jax:
    jit = jax.jit
    fori_loop = jax.lax.fori_loop
    from jax.scipy.linalg import cho_factor, cho_solve, qr, solve_triangular

    def put(arr, inds, vals):
        """Functional interface for array "fancy indexing"

        basically a way to do arr[inds] = vals in a way that plays nice with jit/autodiff.

        Parameters
        ----------
        arr : array-like
            Array to populate
        inds : array-like of int
            Indices to populate
        vals : array-like
            Values to insert

        Returns
        -------
        arr : array-like
            Input array with vals inserted at inds.

        """

        return jax.ops.index_update(arr, inds, vals)


else:
    jit = lambda func, *args, **kwargs: func
    from scipy.linalg import cho_factor, cho_solve, qr, solve_triangular

    def put(arr, inds, vals):
        """Functional interface for array "fancy indexing"

        basically a way to do arr[inds] = vals in a way that plays nice with jit/autodiff.

        Parameters
        ----------
        arr : array-like
            Array to populate
        inds : array-like of int
            Indices to populate
        vals : array-like
            Values to insert

        Returns
        -------
        arr : array-like
            Input array with vals inserted at inds.

        """

        arr[inds] = vals
        return arr

    def fori_loop(lower, upper, body_fun, init_val):
        """Loop from lower to upper, applying body_fun to init_val

        This version is for the numpy backend, for jax backend see jax.lax.fori_loop
        The semantics of ``fori_loop`` are given by this Python implementation::

            def fori_loop(lower, upper, body_fun, init_val):
                val = init_val
                for i in range(lower, upper):
                    val = body_fun(i, val)
                return val

        Parameters
        ----------
        lower : int
            an integer representing the loop index lower bound (inclusive)
        upper : int
            an integer representing the loop index upper bound (exclusive)
        body_fun : callable
            function of type ``(int, a) -> a``.
        init_val : array-like or container
            initial loop carry value of type ``a``

        Returns
        -------
        final_val: array-like or container
            Loop value from the final iteration, of type ``a``.

        """
        val = init_val
        for i in np.arange(lower, upper):
            val = body_fun(i, val)
        return val



import numpy as np
import scipy.linalg
from itertools import permutations, combinations_with_replacement
from termcolor import colored
import warnings

from desc.backend import jnp, put
from desc.utils import equals, issorted, isalmostequal, islinspaced
from desc.io import IOAble
from desc.grid import Grid, LinearGrid, ConcentricGrid, QuadratureGrid
from desc.basis import (
    PowerSeries,
    FourierSeries,
    DoubleFourierSeries,
    FourierZernikeBasis,
)


class Transform(IOAble):
    """Transforms from spectral coefficients to real space values

    Parameters
    ----------
    grid : Grid
        Collocation grid of real space coordinates
    basis : Basis
        Spectral basis of modes
    derivs : int or array-like
        * if an int, order of derivatives needed (default=0)
        * if an array, derivative orders specified explicitly. Shape should be (N,3),
          where each row is one set of partial derivatives [dr, dt, dz]
    rcond : float
         relative cutoff for singular values for inverse fitting
    build : bool
        whether to precompute the transforms now or do it later
    build_pinv : bool
        whether to precompute the pseudoinverse now or do it later
    method : {``'fft'``, ``'direct'``}
        * ``'fft'`` uses fast fourier transforms in the zeta direction, and so must have
          equally spaced toroidal nodes, and the same node pattern on each zeta plane
        * ``'direct'`` uses full matrices and can handle arbitrary node patterns and
          spectral bases.
    """

    _io_attrs_ = ["_grid", "_basis", "_derivatives", "_rcond", "_method"]
    _object_lib_ = {
        "PowerSeries": PowerSeries,
        "FourierSeries": FourierSeries,
        "DoubleFourierSeries": DoubleFourierSeries,
        "FourierZernikeBasis": FourierZernikeBasis,
        "Grid": Grid,
        "LinearGrid": LinearGrid,
        "ConcentricGrid": ConcentricGrid,
        "QuadratureGrid": QuadratureGrid,
    }

    def __init__(
        self,
        grid,
        basis,
        derivs=0,
        rcond=None,
        build=False,
        build_pinv=False,
        method="fft",
    ):

        self._grid = grid
        self._basis = basis
        self._rcond = rcond if rcond is not None else "auto"

        self._derivatives = self._get_derivatives(derivs)
        self._sort_derivatives()
        if method in ["direct", "fft"]:
            self._method = method
        else:
            raise ValueError(
                colored("Unknown Transform method '{}'".format(method), "red")
            )
        self._built = False
        self._built_pinv = False
        self._set_up()
        if build:
            self.build()
        if build_pinv:
            self.build_pinv()

    def _set_up(self):

        if self.method == "fft":
            self._check_inputs_fft(self.grid, self.basis)

        self._matrices = {
            i: {j: {k: {} for k in range(4)} for j in range(4)} for i in range(4)
        }

    def __eq__(self, other):
        """Overloads the == operator

        Parameters
        ----------
        other : Transform
            another Transform object to compare to

        Returns
        -------
        bool
            True if other is a Transform with the same attributes as self
            False otherwise

        """
        if self.__class__ != other.__class__:
            return False
        ignore_keys = ["_built", "_built_pinv", "_matrices"]
        dict1 = {
            key: val for key, val in self.__dict__.items() if key not in ignore_keys
        }
        dict2 = {
            key: val for key, val in other.__dict__.items() if key not in ignore_keys
        }
        return equals(dict1, dict2)

    def _get_derivatives(self, derivs):
        """Get array of derivatives needed for calculating objective function

        Parameters
        ----------
        derivs : int or string
             order of derivatives needed, if an int (Default = 0)
             OR
             array of derivative orders, shape (N,3)
             [dr, dt, dz]

        Returns
        -------
        derivatives : ndarray
            combinations of derivatives needed
            Each row is one set, columns represent the order of derivatives
            for [rho, theta, zeta]

        """
        if isinstance(derivs, int) and derivs >= 0:
            derivatives = np.array([[]])
            combos = combinations_with_replacement(range(derivs + 1), 3)
            for combo in list(combos):
                perms = set(permutations(combo))
                for perm in list(perms):
                    if derivatives.shape[1] == 3:
                        derivatives = np.vstack([derivatives, np.array(perm)])
                    else:
                        derivatives = np.array([perm])
            derivatives = derivatives[
                derivatives.sum(axis=1) <= derivs
            ]  # remove higher orders
        elif np.atleast_1d(derivs).ndim == 1 and len(derivs) == 3:
            derivatives = np.asarray(derivs).reshape((1, 3))
        elif np.atleast_2d(derivs).ndim == 2 and np.atleast_2d(derivs).shape[1] == 3:
            derivatives = np.atleast_2d(derivs)
        else:
            raise NotImplementedError(
                colored(
                    "derivs should be array-like with 3 columns, or a non-negative int",
                    "red",
                )
            )

        return derivatives

    def _sort_derivatives(self):
        """Sorts derivatives"""
        sort_idx = np.lexsort(
            (self.derivatives[:, 0], self.derivatives[:, 1], self.derivatives[:, 2])
        )
        self._derivatives = self.derivatives[sort_idx]

    def _check_inputs_fft(self, grid, basis):
        """helper function to check that inputs are formatted correctly for fft method"""
        if grid.num_nodes == 0 or basis.num_modes == 0:
            # this is the trivial case where we just return all zeros, so it doesn't matter
            self._method = "fft"

        zeta_vals, zeta_cts = np.unique(grid.nodes[:, 2], return_counts=True)

        if len(zeta_vals) % 2 == 0:
            warnings.warn(
                colored(
                    "fft method requires an odd number of toroidal nodes, falling back to direct method",
                    "yellow",
                )
            )
            self._method = "direct"
            return

        if not issorted(grid.nodes[:, 2]):
            warnings.warn(
                colored(
                    "fft method requires nodes to be sorted by toroidal angle in ascending order, falling back to direct method",
                    "yellow",
                )
            )
            self._method = "direct"
            return

        if not isalmostequal(zeta_cts):
            warnings.warn(
                colored(
                    "fft method requires the same number of nodes on each zeta plane, falling back to direct method",
                    "yellow",
                )
            )
            self._method = "direct"
            return

        if len(zeta_vals) > 1:
            if not islinspaced(zeta_vals):
                warnings.warn(
                    colored(
                        "fft method requires nodes to be equally spaced in zeta, falling back to direct method",
                        "yellow",
                    )
                )
                self._method = "direct"
                return

            if not isalmostequal(
                grid.nodes[:, :2].T.reshape((2, zeta_cts[0], -1), order="F")
            ):
                warnings.warn(
                    colored(
                        "fft method requires that node pattern is the same on each zeta plane, falling back to direct method",
                        "yellow",
                    )
                )
                self._method = "direct"
                return

            if not abs((zeta_vals[-1] + zeta_vals[1]) * grid.NFP - 2 * np.pi) < 1e-14:
                warnings.warn(
                    colored(
                        "fft method requires that nodes complete 1 full field period, falling back to direct method",
                        "yellow",
                    )
                )
                self._method = "direct"
                return

        id2 = np.lexsort((basis.modes[:, 1], basis.modes[:, 0], basis.modes[:, 2]))
        if not issorted(id2):
            warnings.warn(
                colored(
                    "fft method requires zernike indices to be sorted by toroidal mode number, falling back to direct method",
                    "yellow",
                )
            )
            self._method = "direct"
            return

        n_vals, n_cts = np.unique(basis.modes[:, 2], return_counts=True)
        if len(n_vals) > 1:
            if not islinspaced(n_vals):
                warnings.warn(
                    colored(
                        "fft method requires the toroidal modes are equally spaced in n, falling back to direct method",
                        "yellow",
                    )
                )
                self._method = "direct"
                return

        if not len(zeta_vals) >= len(n_vals):
            warnings.warn(
                colored(
                    "fft method can not undersample in zeta, num_toroidal_modes={}, num_toroidal_angles={}, falling back to direct method".format(
                        len(n_vals), len(zeta_vals)
                    ),
                    "yellow",
                )
            )
            self._method = "direct"
            return

        self._method = "fft"
        self.lm_modes = np.unique(basis.modes[:, :2], axis=0)
        self.num_lm_modes = self.lm_modes.shape[0]  # number of radial/poloidal modes
        self.num_n_modes = 2 * basis.N + 1  # number of toroidal modes
        self.num_z_nodes = len(zeta_vals)  # number of zeta nodes
        self.N = basis.N  # toroidal resolution of basis
        self.pad_dim = (self.num_z_nodes - 1) // 2 - self.N
        self.dk = basis.NFP * np.arange(-self.N, self.N + 1).reshape((1, -1))
        self.fft_index = np.zeros((basis.num_modes,), dtype=int)
        for k in range(basis.num_modes):
            row = np.where((basis.modes[k, :2] == self.lm_modes).all(axis=1))[0]
            col = np.where(basis.modes[k, 2] == n_vals)[0]
            self.fft_index[k] = self.num_n_modes * row + col
        self.fft_nodes = np.hstack(
            [
                grid.nodes[:, :2][: grid.num_nodes // self.num_z_nodes],
                np.zeros((grid.num_nodes // self.num_z_nodes, 1)),
            ]
        )

    def build(self):
        """Builds the transform matrices for each derivative order"""
        if self.built:
            return

        if self.basis.num_modes == 0:
            self._built = True
            return

        if self.method == "direct":
            for d in self.derivatives:
                self._matrices[d[0]][d[1]][d[2]] = self.basis.evaluate(
                    self.grid.nodes, d
                )

        elif self.method == "fft":
            temp_d = np.hstack(
                [self.derivatives[:, :2], np.zeros((len(self.derivatives), 1))]
            )
            temp_modes = np.hstack([self.lm_modes, np.zeros((self.num_lm_modes, 1))])
            for d in temp_d:
                self.matrices[d[0]][d[1]][d[2]] = self.basis.evaluate(
                    self.fft_nodes, d, modes=temp_modes
                )

        self._built = True

    def build_pinv(self):
        """build pseudoinverse for fitting"""
        if self.built_pinv:
            return
        A = self.basis.evaluate(self.grid.nodes, np.array([0, 0, 0]))
        # for weighted least squares
        A = self.grid.weights[:, np.newaxis] * A
        rcond = None if self.rcond == "auto" else self.rcond
        if A.size:
            self._matrices["pinv"] = scipy.linalg.pinv(A, rcond=rcond)
        else:
            self._matrices["pinv"] = np.zeros_like(A.T)
        self._built_pinv = True

    def transform(self, c, dr=0, dt=0, dz=0):
        """Transform from spectral domain to physical

        Parameters
        ----------
        c : ndarray, shape(num_coeffs,)
            spectral coefficients, indexed to correspond to the spectral basis
        dr : int
            order of radial derivative
        dt : int
            order of poloidal derivative
        dz : int
            order of toroidal derivative

        Returns
        -------
        x : ndarray, shape(num_nodes,)
            array of values of function at node locations
        """
        if not self.built:
            self.build()

        if len(c) == 0:
            return np.zeros(self.grid.num_nodes)

        if self.method == "direct":
            A = self.matrices[dr][dt][dz]
            if isinstance(A, dict):
                raise ValueError(
                    colored("Derivative orders are out of initialized bounds", "red")
                )
            if self.basis.num_modes != c.size:
                raise ValueError(
                    colored(
                        "Coefficients dimension ({}) is incompatible with the number of basis modes({})".format(
                            c.size, self.basis.num_modes
                        ),
                        "red",
                    )
                )
            return jnp.matmul(A, c)

        elif self.method == "fft":
            A = self.matrices[dr][dt][0]
            if isinstance(A, dict):
                raise ValueError(
                    colored("Derivative orders are out of initialized bounds", "red")
                )
            if self.basis.num_modes != c.size:
                raise ValueError(
                    colored(
                        "Coefficients dimension ({}) is incompatible with the number of basis modes({})".format(
                            c.size, self.basis.num_modes
                        ),
                        "red",
                    )
                )

            # reshape coefficients
            c_mtrx = jnp.zeros((self.num_lm_modes * self.num_n_modes,))
            c_mtrx = put(c_mtrx, self.fft_index, c).reshape((-1, self.num_n_modes))

            # differentiate
            c_diff = c_mtrx[:, :: (-1) ** dz] * self.dk ** dz * (-1) ** (dz > 1)

            # re-format in complex notation
            c_real = jnp.pad(
                (self.num_z_nodes / 2)
                * (c_diff[:, self.N + 1 :] - 1j * c_diff[:, self.N - 1 :: -1]),
                ((0, 0), (0, self.pad_dim)),
                mode="constant",
            )
            c_cplx = jnp.hstack(
                (
                    self.num_z_nodes * c_diff[:, self.N, jnp.newaxis],
                    c_real,
                    jnp.fliplr(jnp.conj(c_real)),
                )
            )

            # transform coefficients
            c_fft = jnp.real(jnp.fft.ifft(c_cplx))
            return jnp.matmul(A, c_fft).flatten(order="F")

    def fit(self, x):
        """Transform from physical domain to spectral using weighted least squares fit

        Parameters
        ----------
        x : ndarray, shape(num_nodes,)
            values in real space at coordinates specified by grid

        Returns
        -------
        c : ndarray, shape(num_coeffs,)
            spectral coefficients in basis

        """
        if not self.built_pinv:
            self.build_pinv()
        return jnp.matmul(self.matrices["pinv"], self.grid.weights * x)

    def project(self, y):
        """Project vector y onto basis

        Equivalent to dotting the transpose of the transform matrix into y, but
        somewhat more efficient in some cases by using FFT instead of full transform

        Parameters
        ----------
        y : ndarray
            vector to project. Should be of size (self.grid.num_nodes,)

        Returns
        -------
        b : ndarray
            vector y projected onto basis, shape (self.basis.num_modes)
        """
        if not self.built:
            self.build()

        if self.method == "direct":
            A = self.matrices[0][0][0]
            if self.grid.num_nodes != y.size:
                raise ValueError(
                    colored(
                        "y dimension ({}) is incompatible with the number of grid nodes({})".format(
                            y.size, self.grid.num_nodes
                        ),
                        "red",
                    )
                )
            return jnp.matmul(A.T, y)

        elif self.method == "fft":
            A = self.matrices[0][0][0]
            if self.grid.num_nodes != y.size:
                raise ValueError(
                    colored(
                        "y dimension ({}) is incompatible with the number of grid nodes({})".format(
                            y.size, self.grid.num_nodes
                        ),
                        "red",
                    )
                )
            # this was derived by trial and error, but seems to work correctly
            # there might be a more efficient way...
            a = jnp.fft.fft(A.T @ y.reshape((A.shape[0], -1), order="F"))
            cdn = a[:, 0]
            cr = a[:, 1 : 1 + self.N]
            b = jnp.hstack(
                [-cr.imag[:, ::-1], cdn.real[:, np.newaxis], cr.real]
            ).flatten()[self.fft_index]
            return b

    def change_resolution(self, grid=None, basis=None, build=True, build_pinv=False):
        """Re-builds the matrices with a new grid and basis

        Parameters
        ----------
        grid : Grid
            Collocation grid of real space coordinates
        basis : Basis
            Spectral basis of modes
        build : bool
            whether to recompute matrices now or wait until requested

        """
        if grid is None:
            grid = self.grid
        if basis is None:
            basis = self.basis

        if self.grid != grid:
            self._grid = grid
            self._built = False
            self._built_pinv = False
        if self.basis != basis:
            self._basis = basis
            self._built = False
            self._built_pinv = False
        if self.method == "fft":
            self._check_inputs_fft(self.grid, self.basis)
        if build:
            self.build()
        if build_pinv:
            self.build_pinv()

    @property
    def grid(self):
        """Grid : collocation grid for the transform"""
        if not hasattr(self, "_grid"):
            self._grid = None
        return self._grid

    @grid.setter
    def grid(self, grid):
        if self.grid != grid:
            self._grid = grid
            if self.method == "fft":
                self._check_inputs_fft(self.grid, self.basis)
            if self.built:
                self._built = False
                self.build()
            if self.built_pinv:
                self._built_pinv = False
                self.build_pinv()

    @property
    def basis(self):
        """Basis : spectral basis for the transform"""
        if not hasattr(self, "_basis"):
            self._basis = None
        return self._basis

    @basis.setter
    def basis(self, basis):
        if self.basis != basis:
            self._basis = basis
            if self.method == "fft":
                self._check_inputs_fft(self.grid, self.basis)
            if self.built:
                self._built = False
                self.build()
            if self.built_pinv:
                self._built_pinv = False
                self.build_pinv()

    @property
    def derivatives(self):
        """Set of derivatives the transform can compute

        Returns
        -------
        derivatives : ndarray
            combinations of derivatives needed
            Each row is one set, columns represent the order of derivatives
            for [rho, theta, zeta]
        """
        return self._derivatives

    def change_derivatives(self, derivs, build=True):
        """Changes the order and updates the matrices accordingly

        Doesn't delete any old orders, only adds new ones if not already there

        Parameters
        ----------
        derivs : int or array-like
              * if an int, order of derivatives needed (default=0)
              * if an array, derivative orders specified explicitly.
                shape should be (N,3), where each row is one set of partial derivatives
                [dr, dt, dz]

        build : bool
            whether to build transforms immediately or wait

        """
        new_derivatives = self._get_derivatives(derivs)
        new_not_in_old = (new_derivatives[:, None] == self.derivatives).all(-1).any(-1)
        derivs_to_add = new_derivatives[~new_not_in_old]
        self._derivatives = np.vstack([self.derivatives, derivs_to_add])
        self._sort_derivatives()

        if len(derivs_to_add):
            # if we actually added derivatives and didn't build them, then its not built
            self._built = False
        if build:
            # we don't update self._built here because if it was built before it still is
            # but if it wasn't it still might have unbuilt matrices
            self.build()

    @property
    def matrices(self):
        """dict of ndarray : transform matrices such that x=A*c"""
        if not hasattr(self, "_matrices"):
            self._matrices = {
                i: {j: {k: {} for k in range(4)} for j in range(4)} for i in range(4)
            }
        return self._matrices

    @property
    def num_nodes(self):
        """int : number of nodes in the collocation grid"""
        return self.grid.num_nodes

    @property
    def num_modes(self):
        """int : number of modes in the spectral basis"""
        return self.basis.num_modes

    @property
    def modes(self):
        """ndarray: collocation nodes"""
        return self.grid.nodes

    @property
    def nodes(self):
        """ndarray: spectral mode numbers"""
        return self.basis.nodes

    @property
    def built(self):
        """bool : whether the transform matrices have been built"""
        if not hasattr(self, "_built"):
            self._built = False
        return self._built

    @property
    def built_pinv(self):
        """bool : whether the pseudoinverse matrix has been computed for inverse fitting"""
        if not hasattr(self, "_built_pinv"):
            self._built_pinv = False
        return self._built_pinv

    @property
    def rcond(self):
        """float: reciprocal condition number for inverse transform"""
        if not hasattr(self, "_rcond"):
            self._rcond = "auto"
        return self._rcond

    @property
    def method(self):
        """{``'direct'``, ``'fft'``}: method of computing transform"""
        if not hasattr(self, "_method"):
            self._method = "direct"
        return self._method

    def __repr__(self):
        """string form of the object"""
        return (
            type(self).__name__
            + " at "
            + str(hex(id(self)))
            + " (method={}, basis={}, grid={})".format(
                self.method, repr(self.basis), repr(self.grid)
            )
        )



import numpy as np

from desc.utils import sign
from desc.basis import jacobi


def ptolemy_identity_fwd(m_0, n_0, s, c):
    """Convert from double-angle to double-Fourier form using Ptolemy's identity.

    Converts from the double-angle form:
        s*sin(m*theta-n*phi) + c*cos(m*theta-n*phi)
    to a double Fourier series of the form:
        ss*sin(m*theta)*sin(n*phi) + sc*sin(m*theta)*cos(n*phi) +
        cs*cos(m*theta)*sin(n*phi) + cc*cos(m*theta)*cos(n*phi)
    using Ptolemy's sum and difference formulas.

    Parameters
    ----------
    m_0 : ndarray
        Poloidal mode numbers of the double-angle Fourier basis.
    n_0 : ndarray
        Toroidal mode numbers of the double-angle Fourier basis.
    s : ndarray, shape(surfs,num_modes), optional
        Coefficients of sin(m*theta-n*phi) terms.
        Each row is a separate flux surface.
    c : ndarray, shape(surfs,num_modes), optional
        Coefficients of cos(m*theta-n*phi) terms.
        Each row is a separate flux surface.

    Returns
    -------
    m_1 : ndarray, shape(num_modes,)
        Poloidal mode numbers of the double Fourier basis.
    n_1 : ndarray, shape(num_modes,)
        Toroidal mode numbers of the double Fourier basis.
    x : ndarray, shape(surfs,num_modes,)
        Spectral coefficients in the double Fourier basis.

    """
    s = np.atleast_2d(s)
    c = np.atleast_2d(c)

    M = int(np.max(np.abs(m_0)))
    N = int(np.max(np.abs(n_0)))

    mn_1 = np.array(
        [[m - M, n - N] for m in range(2 * M + 1) for n in range(2 * N + 1)]
    )
    m_1 = mn_1[:, 0]
    n_1 = mn_1[:, 1]
    x = np.zeros((s.shape[0], m_1.size))

    for k in range(len(m_0)):
        # sin(m*theta)*cos(n*phi)
        sin_mn_1 = np.where((mn_1 == [-np.abs(m_0[k]), np.abs(n_0[k])]).all(axis=1))[0][
            0
        ]
        # cos(m*theta)*sin(n*phi)
        sin_mn_2 = np.where((mn_1 == [np.abs(m_0[k]), -np.abs(n_0[k])]).all(axis=1))[0][
            0
        ]
        # cos(m*theta)*cos(n*phi)
        cos_mn_1 = np.where((mn_1 == [np.abs(m_0[k]), np.abs(n_0[k])]).all(axis=1))[0][
            0
        ]
        # sin(m*theta)*sin(n*phi)
        cos_mn_2 = np.where((mn_1 == [-np.abs(m_0[k]), -np.abs(n_0[k])]).all(axis=1))[
            0
        ][0]

        if np.sign(m_0[k]) != 0:
            x[:, sin_mn_1] += s[:, k]
        x[:, cos_mn_1] += c[:, k]
        if np.sign(n_0[k]) > 0:
            x[:, sin_mn_2] -= s[:, k]
            if np.sign(m_0[k]) != 0:
                x[:, cos_mn_2] += c[:, k]
        elif np.sign(n_0[k]) < 0:
            x[:, sin_mn_2] += s[:, k]
            if np.sign(m_0[k]) != 0:
                x[:, cos_mn_2] -= c[:, k]

    return m_1, n_1, x


def ptolemy_identity_rev(m_1, n_1, x):
    """Convert from double-Fourier to double-angle form using Ptolemy's identity.

    Converts from a double Fourier series of the form:
        ss*sin(m*theta)*sin(n*phi) + sc*sin(m*theta)*cos(n*phi) +
        cs*cos(m*theta)*sin(n*phi) + cc*cos(m*theta)*cos(n*phi)
    to the double-angle form:
        s*sin(m*theta-n*phi) + c*cos(m*theta-n*phi)
    using Ptolemy's sum and difference formulas.

    Parameters
    ----------
    m_1 : ndarray, shape(num_modes,)
        Poloidal mode numbers of the double Fourier basis.
    n_1 : ndarray, shape(num_modes,)
        Toroidal mode numbers of the double Fourier basis.
    x : ndarray, shape(surfs,num_modes,)
        Spectral coefficients in the double Fourier basis.

    Returns
    -------
    m_0 : ndarray
        Poloidal mode numbers of the double-angle Fourier basis.
    n_0 : ndarray
        Toroidal mode numbers of the double-angle Fourier basis.
    s : ndarray, shape(surfs,num_modes)
        Coefficients of sin(m*theta-n*phi) terms.
        Each row is a separate flux surface.
    c : ndarray, shape(surfs,num_modes)
        Coefficients of cos(m*theta-n*phi) terms.
        Each row is a separate flux surface.

    """
    x = np.atleast_2d(x)

    M = int(np.max(np.abs(m_1)))
    N = int(np.max(np.abs(n_1)))

    mn_0 = np.array([[m, n - N] for m in range(M + 1) for n in range(2 * N + 1)])
    mn_0 = mn_0[N:, :]
    m_0 = mn_0[:, 0]
    n_0 = mn_0[:, 1]

    s = np.zeros((x.shape[0], m_0.size))
    c = np.zeros_like(s)

    for k in range(len(m_1)):
        # (|m|*theta + |n|*phi)
        idx_pos = np.where((mn_0 == [np.abs(m_1[k]), -np.abs(n_1[k])]).all(axis=1))[0]
        # (|m|*theta - |n|*phi)
        idx_neg = np.where((mn_0 == [np.abs(m_1[k]), np.abs(n_1[k])]).all(axis=1))[0]

        # if m == 0 and n != 0, p = 0; otherwise p = 1
        p = int(bool(m_1[k])) ** int(bool(n_1[k]))

        if sign(m_1[k]) * sign(n_1[k]) < 0:
            # sin_mn terms
            if idx_pos.size:
                s[:, idx_pos[0]] += x[:, k] / (2 ** p)
            if idx_neg.size:
                s[:, idx_neg[0]] += x[:, k] / (2 ** p) * sign(n_1[k])
        else:
            # cos_mn terms
            if idx_pos.size:
                c[:, idx_pos[0]] += x[:, k] / (2 ** p) * sign(n_1[k])
            if idx_neg.size:
                c[:, idx_neg[0]] += x[:, k] / (2 ** p)

    return m_0, n_0, s, c


def fourier_to_zernike(m, n, x_mn, basis):
    """Convert from a double Fourier series to a Fourier-Zernike basis.

    Parameters
    ----------
    m : ndarray, shape(num_modes,)
        Poloidal mode numbers.
    n : ndarray, shape(num_modes,)
        Toroidal mode numbers.
    x_mn : ndarray, shape(surfs,num_modes)
        Spectral coefficients in the double Fourier basis.
        Each row is a separate flux surface, increasing from the magnetic
        axis to the boundary.
    basis : FourierZernikeBasis
        Basis set for x_lmn

    Returns
    -------
    x_lmn : ndarray, shape(num_modes,)
        Fourier-Zernike spectral coefficients.

    """
    x_lmn = np.zeros((basis.num_modes,))
    surfs = x_mn.shape[0]
    rho = np.sqrt(np.linspace(0, 1, surfs))

    for k in range(len(m)):
        idx = np.where((basis.modes[:, 1:] == [m[k], n[k]]).all(axis=1))[0]
        if len(idx):
            A = jacobi(rho, basis.modes[idx, 0], basis.modes[idx, 1])
            c = np.linalg.lstsq(A, x_mn[:, k], rcond=None)[0]
            x_lmn[idx] = c

    return x_lmn


def zernike_to_fourier(x_lmn, basis, rho):
    """Convert from a Fourier-Zernike basis to a double Fourier series.

    Parameters
    ----------
    x_lmn : ndarray, shape(num_modes,)
        Fourier-Zernike spectral coefficients.
    basis : FourierZernikeBasis
        Basis set for x_lmn.
    rho : ndarray
        Radial coordinates of flux surfaces, rho = sqrt(psi).

    Returns
    -------
    m : ndarray, shape(num_modes,)
        Poloidal mode numbers.
    n : ndarray, shape(num_modes,)
        Toroidal mode numbers.
    x_mn : ndarray, shape(surfs,num_modes)
        Spectral coefficients in the double Fourier basis.
        Each row is a separate flux surface, increasing from the magnetic
        axis to the boundary.

    """
    M = basis.M
    N = basis.N

    mn = np.array([[m - M, n - N] for m in range(2 * M + 1) for n in range(2 * N + 1)])
    m = mn[:, 0]
    n = mn[:, 1]

    x_mn = np.zeros((rho.size, m.size))
    for k in range(len(m)):
        idx = np.where((basis.modes[:, 1:] == [m[k], n[k]]).all(axis=1))[0]
        if len(idx):
            A = jacobi(rho, basis.modes[idx, 0], basis.modes[idx, 1])
            x_mn[:, k] = np.matmul(A, x_lmn[idx])

    return m, n, x_mn
